$script:versionId = "1.0013"
# Registry key for defenfer
$script:GPRegPath = "Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Policies\Microsoft\Windows Defender"
$script:PrefRegPath = "Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows Defender"

# Defender settings have several possible name prefixes, according to their CSP path:
$script:settingNamePrefixPolicyCSP = "Windows:Device:Policy.Config.Defender."
$script:settingNamePrefixDefenderCSP = "Windows:Device:Defender.Configuration."
$script:settingNamePrefixes = @($script:settingNamePrefixPolicyCSP, $script:settingNamePrefixDefenderCSP)

$timeLimitForSettingAction = 50 #in milliseconds. actions that take longer than that are logged.

$settingsNamesMapping = @{
    CheckForSignaturesBeforeRunningScan = "CheckForSignaturesBeforeRunningScan";
    CloudBlockLevel = "CloudBlockLevel"; # expected values <CloudBlockLevelType>: { Default, High, HighPlus, ZeroTolerance }
    CloudExtendedTimeout = "CloudExtendedTimeout"; 
    AllowArchiveScanning = "DisableArchiveScanning";
    AllowBehaviorMonitoring = "DisableBehaviorMonitoring";
    DisableCatchupFullScan = "DisableCatchupFullScan";
    DisableCatchupQuickScan = "DisableCatchupQuickScan";
    AllowEmailScanning = "DisableEmailScanning";
    AllowIOAVProtection = "DisableIOAVProtection";
    AllowRealtimeMonitoring = "DisableRealtimeMonitoring";
    AllowFullScanRemovableDriveScanning = "DisableRemovableDriveScanning";
    AllowFullScanOnMappedNetworkDrives = "DisableScanningMappedNetworkDrivesForFullScan";
    AllowScanningNetworkFiles = "DisableScanningNetworkFiles";
    AllowScriptScanning = "DisableScriptScanning";
    EnableLowCpuPriority = "EnableLowCpuPriority";
    EnableNetworkProtection = "EnableNetworkProtection"; # expected values <ASRRuleActionType>: { Disabled, Enabled, AuditMode }
    ExcludedExtensions = "ExclusionExtension";
    ExcludedPaths = "ExclusionPath";
    ExcludedProcesses = "ExclusionProcess";
    AllowCloudProtection = "MAPSReporting";  # expected values <MAPSReportingType>: { Disabled, Basic, Advanced }
    PUAProtection = "PUAProtection"; # expected values <PUAProtectionType>: { Disabled, Enabled, AuditMode }
    DaysToRetainCleanedMalware = "QuarantinePurgeItemsAfterDelay";
    RealTimeScanDirection = "RealTimeScanDirection"; # expected values <ScanDirection>: { Both, Incoming, Outcoming }
    AvgCPULoadFactor = "ScanAvgCPULoadFactor";
    ScanParameter = "ScanParameters"; # expected values <ScanType>: { QuickScan, FullScan }
    ScheduleScanDay = "ScanScheduleDay"; # expected values <Day>
    ScheduleQuickScanTime = "ScanScheduleQuickScanTime";
    ScheduleScanTime = "ScanScheduleTime";
    SignatureUpdateFileSharesSources = "SignatureDefinitionUpdateFileSharesSources";
    SignatureUpdateFallbackOrder = "SignatureFallbackOrder";
    SignatureUpdateInterval = "SignatureUpdateInterval";
    SubmitSamplesConsent = "SubmitSamplesConsent"; # expected values <SubmitSamplesConsentType>: { AlwaysPrompt, SendSafeSamples, NeverSend, SendAllSamples }
    AllowUserUIAccess = "UILockdown";
    PlatformUpdatesChannel = "PlatformRing"; #no cmdlet for this setting
    EngineUpdatesChannel = "EngineRing"; #no cmdlet for this setting
    DisableGradualRelease = "DisableGradualRelease"; #no cmdlet for this setting
    SecurityIntelligenceUpdateChannel = "SignaturesRing"; #no cmdlet for this setting
    AllowOnAccessProtection = "DisableOnAccessProtection"; #no cmdlet for this setting
    DisableLocalAdminMerge = "DisableLocalAdminMerge";
    ThreatSeverityDefaultAction = "ThreatSeverityDefaultAction"; # expected values are list of key-value (action=level) pairs seperated by '|'. Possible actions and levels can be found in 'threatLevelsPrefNamesMapping' and 'threatActions' structs.
}

$script:settingsByCSPPrefix = @{
    $script:settingNamePrefixDefenderCSP = @($settingsNamesMapping.DisableLocalAdminMerge)
}

# holds registry directory names for this modules settings.
# do not use directly, but call $settingsRegMapping[$settingsNamesMapping.$settingSuffix][0]
$regKeyFolder = @{
    Exclusions = "Exclusions";
    MpEngine = "MpEngine";
    NetworkProtection = "Windows Defender Exploit Guard\Network Protection";
    NetworkProtectionNIS = "NIS";
    Quarantine = "Quarantine";
    RealTimeProtection = "Real-Time Protection";
    Scan = "Scan";
    SignatureUpdates = "Signature Updates";
    Spynet = "Spynet";
    Threats = "Threats\ThreatSeverityDefaultAction";
    UXConfiguration = "UX Configuration";
}

# Used for mapping setting name -> information of the registry to hold that setting's value
# Every Setting entry is as follows:
# @( regKey folder, regValue name , regValueType - if doesn't exist, DWORD assumed)
$settingsRegMapping = @{
    $settingsNamesMapping.CheckForSignaturesBeforeRunningScan = @($regKeyFolder.Scan, "CheckForSignaturesBeforeRunningScan");
    $settingsNamesMapping.CloudBlockLevel = @($regKeyFolder.MpEngine, "MpCloudBlockLevel"); 
    $settingsNamesMapping.CloudExtendedTimeout = @($regKeyFolder.MpEngine, "MpBafsExtendedTimeout"); 
    $settingsNamesMapping.AllowArchiveScanning = @($regKeyFolder.Scan, "DisableArchiveScanning");
    $settingsNamesMapping.AllowBehaviorMonitoring = @($regKeyFolder.RealTimeProtection, "DisableBehaviorMonitoring");
    $settingsNamesMapping.DisableCatchupFullScan = @($regKeyFolder.Scan, "DisableCatchupFullScan");
    $settingsNamesMapping.DisableCatchupQuickScan = @($regKeyFolder.Scan, "DisableCatchupQuickScan");
    $settingsNamesMapping.AllowEmailScanning = @($regKeyFolder.Scan, "DisableEmailScanning");
    $settingsNamesMapping.AllowIOAVProtection = @($regKeyFolder.RealTimeProtection, "DisableIOAVProtection");
    $settingsNamesMapping.AllowRealtimeMonitoring = @($regKeyFolder.RealTimeProtection, "DisableRealtimeMonitoring");
    $settingsNamesMapping.AllowFullScanRemovableDriveScanning = @($regKeyFolder.Scan, "DisableRemovableDriveScanning");
    $settingsNamesMapping.AllowFullScanOnMappedNetworkDrives = @($regKeyFolder.Scan, "DisableScanningMappedNetworkDrivesForFullScan");
    $settingsNamesMapping.AllowScanningNetworkFiles = @($regKeyFolder.Scan, "DisableScanningNetworkFiles");
    $settingsNamesMapping.AllowScriptScanning = @($regKeyFolder.RealTimeProtection, "DisableScriptScanning");
    $settingsNamesMapping.EnableLowCpuPriority = @($regKeyFolder.Scan, "LowCpuPriority");
    $settingsNamesMapping.EnableNetworkProtection = @($regKeyFolder.NetworkProtection, "EnableNetworkProtection"); 
    $settingsNamesMapping.AllowCloudProtection = @($regKeyFolder.Spynet, "SpyNetReporting");  
    $settingsNamesMapping.PUAProtection = @("", "PUAProtection"); 
    $settingsNamesMapping.DaysToRetainCleanedMalware = @($regKeyFolder.Quarantine, "PurgeItemsAfterDelay");
    $settingsNamesMapping.RealTimeScanDirection = @($regKeyFolder.RealTimeProtection, "RealTimeScanDirection"); 
    $settingsNamesMapping.AvgCPULoadFactor = @($regKeyFolder.Scan, "AvgCPULoadFactor");
    $settingsNamesMapping.ScanParameter = @($regKeyFolder.Scan, "ScanParameters"); 
    $settingsNamesMapping.ScheduleScanDay = @($regKeyFolder.Scan, "ScheduleDay");
    $settingsNamesMapping.ScheduleQuickScanTime = @($regKeyFolder.Scan, "ScheduleQuickScanTime");
    $settingsNamesMapping.ScheduleScanTime =  @($regKeyFolder.Scan, "ScheduleTime");
    $settingsNamesMapping.SignatureUpdateFileSharesSources = @($regKeyFolder.SignatureUpdates, "DefinitionUpdateFileSharesSources", [Microsoft.Win32.RegistryValueKind]::String);
    $settingsNamesMapping.SignatureUpdateFallbackOrder = @($regKeyFolder.SignatureUpdates, "FallbackOrder", [Microsoft.Win32.RegistryValueKind]::String);
    $settingsNamesMapping.SignatureUpdateInterval = @($regKeyFolder.SignatureUpdates, "SignatureUpdateInterval");
    $settingsNamesMapping.SubmitSamplesConsent = @($regKeyFolder.Spynet, "SubmitSamplesConsent"); 
    $settingsNamesMapping.AllowUserUIAccess = @($regKeyFolder.UXConfiguration, "UILockdown");
    $settingsNamesMapping.AllowOnAccessProtection =  @($regKeyFolder.RealTimeProtection, "DisableOnAccessProtection");
    $settingsNamesMapping.ExcludedExtensions =  @($regKeyFolder.Exclusions, "Extensions",[Microsoft.Win32.RegistryValueKind]::String); 
    $settingsNamesMapping.ExcludedPaths = @($regKeyFolder.Exclusions, "Paths",[Microsoft.Win32.RegistryValueKind]::String);
    $settingsNamesMapping.ExcludedProcesses = @($regKeyFolder.Exclusions, "Processes",[Microsoft.Win32.RegistryValueKind]::String); 
    $settingsNamesMapping.PlatformUpdatesChannel = @("", "PlatformRing");
    $settingsNamesMapping.EngineUpdatesChannel = @("", "EngineRing");
    $settingsNamesMapping.DisableGradualRelease = @($regKeyFolder.MpEngine, "DisableGradualRelease");
    $settingsNamesMapping.SecurityIntelligenceUpdateChannel = @("", "SignaturesRing");
    AllowNetworkProtectionOnWinServer = @($regKeyFolder.NetworkProtection, "AllowNetworkProtectionOnWinServer");
    DisableLocalAdminMerge = @("", "DisableLocalAdminMerge");
    $settingsNamesMapping.ThreatSeverityDefaultAction = @($regKeyFolder.Threats, "");
}

#enum for readable access to $settingsRegMapping's values
Add-Type -TypeDefinition @"
   public enum regInfoIndex
   {
      dir = 0, //specific sub directory (in addition to regPath of all module)
      name = 1, //property name
      type = 2 //value type
   }
"@

#Exclusion values are merge-supported - the actual value to be set in each setting can be composed from multiple policies.
#These settings act different from other settings, and are set and removed differently - Set is done in 'CompletionActions'
#Values will be removed if we got a remove command for them, or if the policy was changed to include a value that was not there before.
$script:mergeSettingsNamesMapping = @(
    $settingsNamesMapping.ExcludedExtensions,
    $settingsNamesMapping.ExcludedPaths,
    $settingsNamesMapping.ExcludedProcesses,
    $settingsNamesMapping.ThreatSeverityDefaultAction
)

#The set action is done by gathering all apply values in 'Set' function into the following arrays, and setting the final value in 'CompletionActions'.
$script:exclusionValuesToBeSet = @{
    $settingsNamesMapping.ExcludedExtensions = @()
    $settingsNamesMapping.ExcludedPaths = @()
    $settingsNamesMapping.ExcludedProcesses = @()
}

$booleanValues = 
    $settingsNamesMapping.CheckForSignaturesBeforeRunningScan,
    $settingsNamesMapping.AllowArchiveScanning,
    $settingsNamesMapping.AllowAutoExclusions,
    $settingsNamesMapping.AllowBehaviorMonitoring,
    $settingsNamesMapping.DisableCatchupFullScan,
    $settingsNamesMapping.DisableCatchupQuickScan,
    $settingsNamesMapping.AllowEmailScanning,
    $settingsNamesMapping.AllowIOAVProtection,
    $settingsNamesMapping.AllowRealtimeMonitoring,
    $settingsNamesMapping.AllowFullScanRemovableDriveScanning,
    $settingsNamesMapping.AllowFullScanOnMappedNetworkDrives,
    $settingsNamesMapping.AllowScanningNetworkFiles,
    $settingsNamesMapping.AllowScriptScanning,
    $settingsNamesMapping.EnableLowCPUPriority,
    $settingsNamesMapping.AllowUserUIAccess,
    $settingsNamesMapping.AllowOnAccessProtection,
    $settingsNamesMapping.DisableLocalAdminMerge

#settings that need toggling of bool value 
$reverseBooleanValues = 
    $settingsNamesMapping.AllowUserUIAccess,
    $settingsNamesMapping.AllowArchiveScanning,
    $settingsNamesMapping.AllowBehaviorMonitoring,
    $settingsNamesMapping.AllowEmailScanning,
    $settingsNamesMapping.AllowIOAVProtection,
    $settingsNamesMapping.AllowRealtimeMonitoring,
    $settingsNamesMapping.AllowFullScanRemovableDriveScanning,
    $settingsNamesMapping.AllowFullScanOnMappedNetworkDrives,
    $settingsNamesMapping.AllowScanningNetworkFiles,
    $settingsNamesMapping.AllowScriptScanning,
    $settingsNamesMapping.AllowOnAccessProtection

#AV settings that do not have cmdlets configured for them. Any action for these settings needs to be done through registry.
$script:registryOnlySettings = @(
    $settingsNamesMapping.AllowOnAccessProtection,
    $settingsNamesMapping.PlatformUpdatesChannel,
    $settingsNamesMapping.EngineUpdatesChannel,
    $settingsNamesMapping.DisableGradualRelease,
    $settingsNamesMapping.SecurityIntelligenceUpdateChannel,
    $settingsNamesMapping.DisableLocalAdminMerge
)

# taken from TP_REGISTRY_KEYS and TP_CDP_REGISTRY_KEYS from https://microsoft.visualstudio.com/Windows%20Defender/_git/WD.Signatures?path=/amcore/signature/Source/mavsigs/infrastructure/packersigs/smartscancfg.inc&version=GBdevelop
$script:settingBlockedByTamperProtection = 
    $settingsNamesMapping.AllowArchiveScanning,
    $settingsNamesMapping.AllowScanningNetworkFiles,
    $settingsNamesMapping.AllowCloudProtection,
    $settingsNamesMapping.AllowOnAccessProtection,
    $settingsNamesMapping.AllowRealtimeMonitoring,
    $settingsNamesMapping.AllowBehaviorMonitoring,
    $settingsNamesMapping.AllowIOAVProtection,
    $settingsNamesMapping.AllowScriptScanning,
    $settingsNamesMapping.ThreatSeverityDefaultAction


$allowCloudProtectionActiveValues = "1", "2" #both Basic (=1) and Advanced (=2) are considered as Active. "0" as inactive.

$script:threatLevelsPrefNamesMapping = @{
    "1" = "LowThreatDefaultAction";
    "2" = "ModerateThreatDefaultAction";
    "4" = "HighThreatDefaultAction";
    "5" = "SevereThreatDefaultAction";
}

$script:supportedThreatActions = @(
    1, # Clean. Service tries to recover files and try to disinfect.
    2, # Quarantine. Moves files to quarantine.
    3, # Remove. Removes files from system.
    6, # Allow. Allows file/does none of the above actions.
    8, # User defined. Requires user to make a decision on which action to take.
    10 # Block. Blocks file execution.
)

$script:detectedThreatsActionUpdated = $false

$script:threatActionPolicyCounter = 0 #maps setting name to counter for policies using it. Used to get and report errors according to policies.
$script:requestedThreatActions = @{} #maps the requested threats by levels to the action  @{threatlevel = @(action,@(groups),@(HResult,ErroStr))}. Values are stored in Set and applied on CompletionActions

# threatActionFirstErrorInfo - error information, in case there is no specific error in $script:requestedThreatActions
$threatActionFirstErrorInfo = @{}

$script:conflictUtilsModule = $null

#Dummy value for TriggerGPRefresh - adding and removing it from MpPreference
$script:dummyValue = "bfbfb459-07db-4fd1-a739-42c533f1119c-SIAM-7adcea70-53c6-43c6-86f0-83d6d5c00459"

if (-not ([System.Management.Automation.PSTypeName]'firstErrorIndex').Type)
{
    Add-Type -TypeDefinition @"
        public enum firstErrorIndex {
            HR,
            errStr
        }
"@
}

#used to detect conflict and report accordingly for Actions for detected threats
$script:conflictingThreatGroups = @()

# index enum (for usage of $script:requestedThreatActions)
Add-Type -TypeDefinition @"
    public enum threatInfoIndex {
        action,
        groups,
        FirstErrorInfo
    }
"@ 

# Supported on various Defender's versions
$notApplicableOn2012r2 = @(
   $settingsNamesMapping.AllowUserUIAccess 
)
    
function IsSupportedParam([string] $name) 
{
    $validPrefix = GetValidPrefix -fullName $name

    if([string]::IsNullOrEmpty($validPrefix))
    {
        #none of the possible prefixes matches the setting name - setting not recognized
        return $false
    }

    $nameSuffix = $name.Replace($validPrefix, "")
    
    if(!($settingsNamesMapping.ContainsKey($nameSuffix)))
    {
        #no such setting name suffix in this module
        return $false
    }

    if($script:settingsByCSPPrefix.ContainsKey($validPrefix) -xor ($script:settingsByCSPPrefix.($script:settingNamePrefixDefenderCSP)).contains($settingsNamesMapping.$nameSuffix))
    {
        #case of name with settingNamePrefixPolicyCSP but it should be with a settingNamePrefixDefenderCSP or vice versa
        return $false
    }

    return $true
}

function Setup()
{
    LogInfo (-join("AntiVirus setup (", $script:versionId, ")"))
 
    #Import ConflictResolutionUtils
    try{
        $script:conflictUtilsModule = Import-Module -Name (Join-Path $global:WorkingFolder "ConflictResolutionUtils.psm1") -Force -AsCustomObject -PassThru -ErrorAction Stop
    } catch {
        LogWarning (-join("AV setup: failed to import ConflictResolutionUtils module. HResult:",$_.Exception.HResult," error message:", $_.Exception.Message))
    }
}

function ResetToDefault([string] $name, [string] $value, [ref] $errorCode, [ref] $errorString)
{
    $errorCode.Value = [HResult]::S_OK.value__
    $errorString.Value = ""
    $local:nameSuffix = ""
    $local:settingPrefix = ""

    try {
        GetNameSuffixAndSettingPrefix $local:name ([ref] $nameSuffix) ([ref] $settingPrefix)
        if ($settingPrefix -ne "")
        {
            $nameSuffix = $nameSuffix.Replace($settingPrefix + ".", "")
        }
        LogInfo (-join("AV remove setting ",$nameSuffix,",",$value))

        #invalid setting name received
        if(!($settingsNamesMapping.ContainsKey($local:nameSuffix)))
        {
            $errorCode.Value = [HResult]::E_NOINTERFACE.value__
            $errorString.Value = WriteErrorSettingIsNotSupported -funcName "ResetToDefault" -name $local:nameSuffix
            return
        }
    
        $startTimeMilli = GetCurrentTimeInMillis

        switch ($settingsNamesMapping.$nameSuffix)
        {
            {$_ -in $exclusionValuesToBeSet.Keys} { RemoveExclusions $_ $local:value ; break }
            {$_ -eq $settingsNamesMapping.ThreatSeverityDefaultAction} 
            { 
                $script:detectedThreatsActionUpdated = $true
                LogInfo (-join("AV: Remove for ",$local:nameSuffix," will happen in CompletionActions."))
                break 
            }
            {$settingsNamesMapping.ContainsValue($_)} # all "special" handling of settings was done in previous cases. Now handle all the "regular" settings (update registry)
            {
                Remove-ItemProperty -Path (Join-Path $script:GPRegPath $settingsRegMapping[$settingsNamesMapping.$local:nameSuffix][([regInfoIndex]::dir)]) -Name $settingsRegMapping[$settingsNamesMapping.$local:nameSuffix][([regInfoIndex]::name)] -ErrorAction SilentlyContinue #if item doesn't exits, continue.
                #workaround for defender issue #39416943 - remove also from pref hive (settings are also set there by cmdlets)
                ResetPrefValue -settingNameSuffix $nameSuffix
            }
        }
    } catch {
        $errorMessage = $_.Exception.Message
        $errorCode.Value = $_.Exception.HResult
        $msg = -join("AV remove failure for ",$nameSuffix,". HResult:",$_.Exception.HResult," error message:",$errorMessage)
        $errorString.Value = $msg
        LogWarning $msg
    }

    $totalTimeMilli = (GetCurrentTimeInMillis) - $startTimeMilli
    if($totalTimeMilli -gt $timeLimitForSettingAction)
    {
        LogInfo (-join("AV: Remove for ",$nameSuffix," took ", $totalTimeMilli," ms."))
    }
}

function Set([string] $name, [string] $value, [ref] $errorCode, [ref] $errorString)
{
    $errorCode.Value = [HResult]::S_OK.value__
    $errorString.Value = ""
    
    $local:nameSuffix = ""
    $local:settingPrefix = ""
    

    try {
        GetNameSuffixAndSettingPrefix $local:name ([ref] $nameSuffix) ([ref] $settingPrefix)
        LogInfo (-join("AV set value ",$local:nameSuffix,", ",$local:value))
        
        #invalid setting name received
        if(!($settingsNamesMapping.ContainsKey($local:nameSuffix)))
        {
            $errorCode.Value = [HResult]::E_NOINTERFACE.value__
            $errorString.Value = WriteErrorSettingIsNotSupported -funcName "Set" -name $local:nameSuffix
            return
        }

        $startTimeMilli = GetCurrentTimeInMillis
        $settingNameValueFromeKey = $settingsNamesMapping.($local:nameSuffix)
         
        switch ($local:settingNameValueFromeKey)
        {
            # we intentionally don't use invoke expression to avoid code injection in case the policy is malicious
            {$_ -in $script:exclusionValuesToBeSet.Keys}
            {
                foreach ($exclusionValue in $local:value.Split("|")) {
                    if($local:exclusionValue) #filtering out empty values
                    {
                        $local:exclusionValue = $local:exclusionValue.Trim() #remove spaces, and check value is not empty after removing it
                        if($local:exclusionValue -and ($local:exclusionValue -notin $script:exclusionValuesToBeSet.$_))
                        {
                            $script:exclusionValuesToBeSet.$_ += $local:exclusionValue #will be set in CompletionActions
                        }
                    }
                }
                break
            }
            {$_ -eq $settingsNamesMapping.ThreatSeverityDefaultAction} 
            {
                $script:detectedThreatsActionUpdated = $true
                $script:threatActionPolicyCounter++  #setting a new group of threat actions  
                #add action to the "to be applied" map, because we might have several policies to apply.
                $threatActionsDict = $script:conflictUtilsModule."ParseSettingInfo"($local:value) # @{} level, action
                $script:conflictUtilsModule."StoreSettingToRequested"($local:threatActionsDict, $script:threatActionPolicyCounter, $script:requestedThreatActions, [ref]$script:conflictingThreatGroups)

                $validErrorMsg = "" #need to initialize to pass as reference
                if(!(IsThreatLevelsAndActionsValid $threatActionsDict ([ref] $validErrorMsg))) #check that all threat levels and actions we recieved are valid.
                {
                    $errorCode.Value = [HResult]::E_INVALIDARG.value__
                    $errorString.Value = $validErrorMsg
                    return 
                }

                break
            }   
            {$settingsNamesMapping.ContainsValue($_)} # all "special" handling of settings was done in previous cases. Now handle all the "regular" settings (update registry)
            { 
                if($_ -in $reverseBooleanValues)
                {
                    $newValue = 1 - $local:value # toggle value 
                    LogInfo (-join("Set for ",$local:nameSuffix," - toggling expected value from ",$local:value," to ",$local:newValue))
                }
                else 
                {
                    $newValue = $local:value
                }

                #value type of reg value is DWORD, unless otherwise stated in $settingsRegMapping
                if(($settingsRegMapping[$_]).Count -gt 2)
                {
                    $valueType = $settingsRegMapping[$_][([regInfoIndex]::type)]
                } 
                else
                {
                    $valueType = [Microsoft.Win32.RegistryValueKind]::DWord
                }

                SetSingleValueToRegistry -regPath (Join-Path $script:GPRegPath $settingsRegMapping[$_][([regInfoIndex]::dir)]) -name $settingsRegMapping[$_][([regInfoIndex]::name)] -type $valueType -value $newValue
                # workaround for defender issue #39416943 - set also to pref hive (by cmdlets)
                # in case of reversed boolean settings (enableX in mem vs disableX in cmdlet or vice-versa), the toggled value is passed to SetPrefValue
                #TP2 - marker for tests - don't remove!
                SetPrefValue -settingNameSuffix $nameSuffix -newValue $newValue
            }
        }
    } catch {
        $errorCode.Value = $_.Exception.HResult
        $errorString.Value = $_.Exception.Message
        LogWarning (-join("AV::Set failure for ",$local:nameSuffix,", value:",$local:value,". HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
    }

    $totalTimeMilli = (GetCurrentTimeInMillis) - $local:startTimeMilli
    if($local:totalTimeMilli -gt $timeLimitForSettingAction)
    {
        LogInfo (-join("AV: Set for ",$local:nameSuffix," took ", $local:totalTimeMilli," ms."))
    }
}

function VerifyAssignment([string] $name, [string] $value, [ref] $enforceSucceeded, [ref] $errorCode, [ref] $errorString)
{
    $local:nameSuffix = ""
    $local:settingPrefix = ""
    $local:hresult = $null
    try
    {    
        GetNameSuffixAndSettingPrefix $name ([ref] $nameSuffix) ([ref] $settingPrefix)
        LogInfo "AV VerAssign $local:nameSuffix,$value"

        # Check if setting supported
        if(!($settingsNamesMapping.ContainsKey($local:nameSuffix)))
        {
            $errorCode.Value = [HResult]::E_NOINTERFACE.value__
            $errorString.Value = (WriteErrorSettingIsNotSupported -funcName "VerifyAssignment" -name $local:nameSuffix)
            $enforceSucceeded.Value = $false
            return [HResult]::E_NOINTERFACE
        }

        $local:startTimeMilli = GetCurrentTimeInMillis

        $local:settingHasCmdlet = !($script:registryOnlySettings.Contains($settingsNamesMapping.$nameSuffix))
      
        $local:expectedValue = $null
        $local:settingCurrentValue = $null
        $local:isExclusions = $false
        $local:isThreatActions = $false


        if ($local:settingHasCmdlet)
        {
            $local:expectedValue = GetValueInCmdletFormat $nameSuffix $value ([ref] $local:isExclusions) ([ref] $local:isThreatActions) $errorCode $errorString
          
            if ($null -eq $local:expectedValue) #expected value should never be null since we got a value for this setting if we got here
            {
                $enforceSucceeded.Value = $false
                $msg = (-join("AV::VerifyAssignment failure for ",$local:nameSuffix,", value:",$local:value,". Expected value is null."))
                if ([string]::IsNullOrEmpty($errorString.Value)) {
                    $errorString.Value = $msg
                } 
                else {
                    $errorString.Value += ". " + $msg
                }
                LogWarning $msg
                $errorCode.value = [HResult]::E_FAIL.value__
                $local:hresult = [HResult]::E_FAIL
            }
            else
            {
                $local:hresultCurrentValueFromMpPref = $null
                $local:settingCurrentValue = GetSettingCurrentValueFromMpPref $nameSuffix $errorCode $errorString ([ref] $local:hresultCurrentValueFromMpPref)
                if ($null -eq $local:settingCurrentValue)
                {   
                    # VerifyAssignment is called after "Set", value should not be empty
                    # We don't have an explicit fail reason, so we don't set errorCode/errorString. Hopefully we have a fail reason for Set.
                    $enforceSucceeded.Value = $false
                    $local:hresult = $local:hresultCurrentValueFromMpPref
                }                     
            }
        }
        else
        {
            $local:expectedValue = GetValueInRegistryFormat $nameSuffix $value ([ref] $local:isExclusions) ([ref] $local:isThreatActions)
            $local:settingCurrentValue = GetSettingCurrentValueFromGPRegistry $nameSuffix $isExclusions
        }

        # This means the flow before didn't set hresult - so we can to continue
        if ($null -eq $local:hresult)
        {
            if($local:isThreatActions)
            {
                $enforceSucceeded.Value = ValidateThreatsAssignment $local:settingCurrentValue $local:expectedValue $errorCode $errorString
                if($errorString.Value -match "conflictUtilsModule") #if utils module was not loaded, we should return E_FAIL since validation failed.
                {
                    $local:hresult = [HResult]::E_Fail
                }
            }
            else
            {
                $enforceSucceeded.Value = (GetEnforceSucceeded $local:settingCurrentValue $local:expectedValue)
            }

            if((!($enforceSucceeded.Value)) -and ($errorCode.Value -ne [HResult]::E_INTERNAL_CONFLICT.value__)) #in case of conflict we expect values will not match
            {
                LogInfo (-join("AV:VerAssign returns false for ",$nameSuffix," (current:",(PrepareValueToPrint $local:settingCurrentValue),", expected:",(PrepareValueToPrint $local:expectedValue), ")"))
            }

            # When TP is on it prevents some setting from being set, by design. We suppress the error in this case to have parity with intune agent.
            # This is temporary untill a holistic solution for all mgmt tools will be established.
            # This might hide actuall set errors (because we override the report of errors that happend during 'set' operation in case the setting value is the same as the requested value). 
            # Such errors will be exposed if TP will be set to off.
            if(!($enforceSucceeded.Value) -and (IsTamperProtectionActiveOnSetting $local:nameSuffix) -and ($errorString.Value -eq ""))
            {
                $enforceSucceeded.Value = $true
                LogWarning ("TamperProtection is activated. Value of setting $nameSuffix is protected - reporting success.")
            }
        }

        $local:totalTimeMilli = (GetCurrentTimeInMillis) - $local:startTimeMilli
        if($local:totalTimeMilli -gt $timeLimitForSettingAction)
        {
            LogInfo "AV:VerAssign for $nameSuffix took $totalTimeMilli ms."
        }
    }
    catch 
    {
        $errorString.Value = $_.Exception.Message
        $errorCode.Value = $_.Exception.HResult 
        LogWarning "AV::VerifyAssignment failure. Setting $local:nameSuffix HR:$($_.Exception.HResult) Message:$($_.Exception.Message)"
        $local:hresult = [HResult]::E_Fail
    }
    
    if ($null -eq $local:hresult)
    {
        $local:hresult = [HResult]::S_OK
    }

    return $local:hresult
}

function VerifyRemoval([string] $name, [string] $value, [ref] $enforceSucceeded, [ref] $errorCode, [ref] $errorString)
{
    $errorCode.Value = [HResult]::S_OK.value__
    $errorString.Value = ""
    $local:nameSuffix = ""
    $local:settingPrefix = ""

    try {
        GetNameSuffixAndSettingPrefix $local:name ([ref] $nameSuffix) ([ref] $settingPrefix)

        if ($local:settingPrefix -ne "")
        {
            $local:nameSuffix = $local:nameSuffix.Replace($local:settingPrefix + ".", "")
        }
        LogInfo (-join("AV VerRemov ",$local:nameSuffix, ",",$local:value))

        if(!($settingsNamesMapping.ContainsKey($local:nameSuffix)))
        {
            $errorCode.Value = [HResult]::E_NOINTERFACE.value__
            $errorString.Value = (WriteErrorSettingIsNotSupported -funcName "VerRemov" -name $local:nameSuffix)
            $enforceSucceeded.Value = $false
            return [HResult]::E_NOINTERFACE
        }

        $startTimeMilli = GetCurrentTimeInMillis

        $currentRegItem = $null

        switch ($settingsNamesMapping.$nameSuffix)
        {
            {$_ -in $script:exclusionValuesToBeSet.Keys} {$enforceSucceeded.Value = IsExlucsionValueRemoved $nameSuffix $value; return [HResult]::S_OK }
            {$settingsNamesMapping.ThreatSeverityDefaultAction} { $enforceSucceeded.Value = IsThreatValueRemoved $value; return [HResult]::S_OK }    
            {$settingsNamesMapping.ContainsValue($_)} # all "special" handling of settings was done in previous cases. Now handle all the "regular" settings 
            {
                $currentRegItem = (GetSingleCurrentValueFromRegistry -nameSuffix $nameSuffix -registryPath (Join-Path $script:GPRegPath $settingsRegMapping[$settingsNamesMapping.$local:nameSuffix][([regInfoIndex]::dir)]))
                if($local:settingHasCmdlet)
                {
                    $currentPrefItem = (GetSingleCurrentValueFromRegistry -nameSuffix $nameSuffix -registryPath (Join-Path $script:PrefRegPath $settingsRegMapping[$settingsNamesMapping.$local:nameSuffix][([regInfoIndex]::dir)]))
                }
            }
        }

        $GPHiveRemoved = ($null -eq $local:currentRegItem) -or ($local:currentRegItem.Length -eq 0)
        $PrefHiveRemoved = $true

        if($local:settingHasCmdlet) 
        {
            $PrefHiveRemoved = ($null -eq $local:currentPrefItem) -or ($local:currentPrefItem.Length -eq 0)
        }

        if ($local:GPHiveRemoved -and $local:PrefHiveRemoved)
        {
            $enforceSucceeded.Value = $true
        } 
        else
        {        
            $fullRegStr = ""
            if(!$local:GPHiveRemoved)
            {
                $local:fullRegStr += (-join("GP:",$local:currentRegItem))
            }
            if(!$local:PrefHiveRemoved)
            {
                if(![string]::IsNullOrEmpty($local:fullRegStr))
                {
                    $local:fullRegStr += ","
                }
                $local:fullRegStr += (-join("Pref:",$local:currentPrefItem))
            }

            LogInfo (-join("AV:VerifyRemoval - expected empty hives that have values:",$local:fullRegStr))
            $enforceSucceeded.Value = $false    
        }

        $totalTimeMilli = (GetCurrentTimeInMillis) - $local:startTimeMilli
        if($totalTimeMilli -gt $timeLimitForSettingAction)
        {
            LogInfo (-join("AV: VerRemov for ",$nameSuffix," took ", $totalTimeMilli," ms."))
        }

    } catch {
        $errorString.Value = $_.Exception.Message
        $errorCode.Value = $_.Exception.HResult 
        LogWarning (-join("AV::VerifyRemoval failure. Setting ",$nameSuffix," HR:",$_.Exception.HResult," msg:",$_.Exception.Message))
        return [HResult]::E_Fail
    }

    return [HResult]::S_OK
}

function IsApplicable([string] $name, [string] $value)
{
    $isSettingApplicable = $true
    
    if ($global:OSVersion -match $global:2012BuildStr)
    {
        # At the moment only 2012R2 has NA settings, so no need to get the name for other OS versions.
        # If more versions have NA settings, getting the name should be done in advance.
        $nameSuffix = GetNameSuffix -fullName $name 
        $settingPrefix = GetSettingPrefix($nameSuffix)
        if ($settingPrefix -ne "")
        {
            $nameSuffix = $nameSuffix.Replace($settingPrefix + ".", "")
        }

        if($settingsNamesMapping.$nameSuffix -in $notApplicableOn2012r2)
        {
            $isSettingApplicable = $false
            LogInfo (-join("AV:IsApplicable setting ",$nameSuffix," is N/A for platform"))
        }
    }

    return $isSettingApplicable
}

function HandleDetectedThreatsAction()
{
    # remove exising values in GP hive only if there is any policy that has DetectedThreatsActions 
    # in them (so we won't touch user's other configuration)
    if($script:detectedThreatsActionUpdated) 
    {
        RemoveThreatActions 
    }

    # will set only if values were requested (and saved in $script:requestedThreatActions)
    SetActionForThreats
}

function CompletionActions() 
{
    LogInfo "AV: closing loose ends." 
    $startTimeMilli = GetCurrentTimeInMillis
    
    # used to set and remove actionsForDetectedThreats values, since we don't get a remove request from MEM if a single value was deleted from a policy. Try/catch in function
    # we only delete values that were in GP hive, not all values from pref hive
    HandleDetectedThreatsAction

    try
    {
        #used to set and remove exclusion values, since we don't get a remove request from MEM if a single value was deleted from a policy
        SetExclusionsSettings 
    } catch {
        LogWarning (-join("AV::CompletionActions failure for SetExclusionsSettings" ,". HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
    }
    
    $totalTimeMilli = (GetCurrentTimeInMillis) - $startTimeMilli
    LogInfo (-join("AV: CompletionActions took ", $totalTimeMilli," ms."))
}

function IsMergeSupported ([string] $name)
{
    $nameSuffix = GetNameSuffix -fullName $name 
    return ($settingsNamesMapping.$nameSuffix -in $script:mergeSettingsNamesMapping)
}

 ####################################################### utility functions: ####################################################### 
 function SetupNWProtectionForDLServer ()
 {
    LogInfo "AV: SetupNWProtectionForDLServer DL Server - setting AllowNetworkProtectionOnWinServer,AllowNetworkProtectionDownLevel,AllowDatagramProcessingOnWinServer to 1"
    SetSingleValueToRegistry -regPath (Join-Path $script:GPRegPath ($settingsRegMapping.AllowNetworkProtectionOnWinServer)[([regInfoIndex]::dir)]) -name ($settingsRegMapping.AllowNetworkProtectionOnWinServer)[([regInfoIndex]::name)] -type ([Microsoft.Win32.RegistryValueKind]::DWord) -value 1
    Set-MpPreference -AllowNetworkProtectionOnWinServer 1
    Set-MpPreference -AllowNetworkProtectionDownLevel 1
    Set-MpPreference -AllowDatagramProcessingOnWinServer 1
}

 function IsDLServer()
{
    if(($global:OSVersion -match $global:2012BuildStr) -or ($global:OSVersion -match $global:2016BuildStr))
    {
        return $true
    }
    return $false
 }

 function GetCurrentValueOfRegistryDir([string] $nameSuffix)
 {
    $value = $null

    $registryItemObj = Get-Item (GetPathToRegItem $local:nameSuffix $true) -ErrorAction SilentlyContinue
    if($registryItemObj)
    {
        $local:value = $registryItemObj.Property
    }

    return $value
 }

function SetSingleValueToRegistry([string] $regPath, [string] $name, [Microsoft.Win32.RegistryValueKind] $type, [string] $value)
{
    if(!(Test-Path $regPath)) 
    {
        LogInfo (-join("AV SetSingleValueToRegistry: reg key ",$regPath, " doesn't exist. creating."))
        New-Item -Path $regPath -Force | Out-Null
    }

    LogInfo (-join("AV Set-ItemProperty: reg key ", $regPath, " name:", $name, " value:",$value))
    Set-ItemProperty -Path $regPath -Name $name -Type $type -Value $value -ErrorAction Stop #SilentlyContinue | Out-Null 
}

function GetSingleCurrentValueFromRegistry([string] $nameSuffix)
{  
    $registryItemObj = Get-ItemProperty (GetPathToKeyDir $local:nameSuffix $true) (GetRegValueName $local:nameSuffix) -ErrorAction SilentlyContinue
    if($registryItemObj)
    {
        return $registryItemObj.((GetRegValueName $local:nameSuffix))
    }
    return $null
}

function WriteErrorSettingIsNotSupported([string] $funcName, [string] $suffixName)
{
    $msg = -join("AV:",$funcName," Setting is not supported ",$suffixName)
    LogInfo $msg
    return $msg
}

function IsTamperProtectionActive()
{  
    #TP1 - marker for tests - don't remove!

    $TPInfo = $null

    try 
    {
        $TPInfo = Get-MpComputerStatus
    } 
    catch 
    {
        LogWarning ( -join("Exception on AV::IsTamperProtectionActive - HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
        return $false
    }

    return ($TPInfo.IsTamperProtected)
}

function GetSettingPrefix([string] $value)
{
    if (!($value -contains "."))
    {
        return $value.Split(".")[0]
    }
    
    return ""
}

function CheckEmptyExclusionValue([string] $settingSuffix, [string] $registryPath)
{
    $hive = "Pref"
    if($registryPath -match "Policies")
    {
        $hive = "GP"
    }

    $regItem = Get-Item -Path $local:registryPath
    if($local:regItem.ValueCount -ne 0)
    {
        foreach($exclusion in $local:regItem.Property)
        {
            $exclusionRequestedForApply = ($script:exclusionValuesToBeSet[$settingsNamesMapping.$local:settingSuffix]).Contains($local:exclusion)
            if(!$exclusionRequestedForApply)
            {
                LogInfo (-join("AV IsExlucsionValueRemoved: setting:",$local:settingSuffix," value ",$local:exclusion," not removed from ", $hive ," hive."))
                return $false
            }
        }
    }
    return $true
}

function IsExlucsionValueRemoved([string]$settingSuffix, [string]$value) 
{
    $result = $true
    $pathToGPRegItem = (GetPathToRegItem $settingSuffix $true)
    $pathToPrefRegItem = (Join-Path $script:PrefRegPath $settingsRegMapping[$settingsNamesMapping.$settingSuffix][([regInfoIndex]::dir)])

    if([string]::IsNullOrEmpty($local:value)) #If value is null need to remove all values
    {
        $local:result = ((CheckEmptyExclusionValue $local:settingSuffix $local:pathToGPRegItem) -and (CheckEmptyExclusionValue $local:settingSuffix $local:pathToPrefRegItem))
    }

    foreach($exclusion in (GetExclusionsSettingValueAsArray $local:value))
    {
        #if the exclusion was later re-applied, we can assume it was removed successfully
        $exclusionRequestedForApply = ($script:exclusionValuesToBeSet[$settingsNamesMapping.$local:settingSuffix]).Contains($local:exclusion)

        $GPRegValueExists = (((Get-ItemProperty -Path $local:pathToGPRegItem).$local:exclusion).Count -ne 0)
        $prefRegValueExists = (((Get-ItemProperty -Path $local:pathToPrefRegItem).$local:exclusion).Count -ne 0)
        if (!$local:exclusionRequestedForApply) #value was not requested to be applied by another policy
        {
            if($local:GPRegValueExists)
            {
                LogInfo (-join("AV IsExlucsionValueRemoved: setting:",$local:settingSuffix," value ",$local:exclusion," not removed from GP hive."))
                $local:result = $false
            }
            if($local:prefRegValueExists)
            {
                LogInfo (-join("AV IsExlucsionValueRemoved: setting:",$local:settingSuffix," value ",$local:exclusion," not removed from Pref hive."))
                $local:result = $false
            }
        }
    }
    return $local:result
}

function IsThreatValueRemoved ([string]$value) 
{
    $result = $true
    $threatLevelsRemoved =  ($value.Split("|") | Where-Object {$_})
    $GPKeyPath = (GetPathToRegItem $settingsNamesMapping.ThreatSeverityDefaultAction $true)
    $prefKeyPath = (Join-Path $script:PrefRegPath $settingsRegMapping[$settingsNamesMapping.ThreatSeverityDefaultAction][([regInfoIndex]::dir)])
    $threatsGPRegValues = Get-ItemProperty -Path $local:GPKeyPath
    $threatsPrefRegValues = Get-ItemProperty -Path $local:prefKeyPath
    foreach($removedLevel in $local:threatLevelsRemoved)
    {
        $level,$action = $local:removedLevel.Split("=")

        if(!($script:requestedThreatActions.ContainsKey($level))) #value was not requested to be applied by another policy
        {
            if($local:threatsGPRegValues.$level) #if value exsits in registry -> report failure
            {
                LogInfo (-join("AV:IsThreatValueRemoved: level:",$local:level,", value: ",$local:action," was not removed from GP hive."))
                $local:result = $false 
            }
            if($local:threatsPrefRegValues.$level) #if value exsits in registry -> report failure
            {
                LogInfo (-join("AV:IsThreatValueRemoved: level:",$local:level,", value: ",$local:action," was not removed from Pref hive."))
                $local:result = $false 
            }
        }
    }
    return $local:result
}

# use GetPathToKeyDir funcation to get the path to the directory of the setting in the resitry.
# if $addRegistryPrefix is true, the "Registry" prefix is added.
# example: for setting ExcludedExtensions
# result will be: 
# if $addRegistryPrefix is false: "HKEY_LOCAL_MACHINE\SOFTWARE\Policies\Microsoft\Windows Defender\Exclusions"
# if $addRegistryPrefix is true: "Registy:HKEY_LOCAL_MACHINE\SOFTWARE\Policies\Microsoft\Windows Defender\Exclusions"
function GetPathToKeyDir([string]$nameSuffix, [bool]$addResitryPrefix = $false)
{
    $path = (Join-Path $script:GPRegPath $settingsRegMapping[$settingsNamesMapping.$local:nameSuffix][([regInfoIndex]::dir)])

    if($addResitryPrefix)
    {
       $path = (-join($script:RegPrefix,$local:path))
    }

    return $path
}

# use GetRegValueName funcation to get the name that represents the value of the setting in the resitry.
# example: for setting ExcludedExtensions
# result will be: "Extensions"
function GetRegValueName([string]$settingSuffix)
{
    return $settingsRegMapping[$settingsNamesMapping.$local:settingSuffix][([regInfoIndex]::name)]
}

# use GetPathToRegItem funcation to get the full path to the name that represents the value of the setting in the resitry.
# example: for setting ExcludedExtensions
# if $addRegPrefix is false: "HKEY_LOCAL_MACHINE\SOFTWARE\Policies\Microsoft\Windows Defender\Exclusions\Extensions"
# if $addRegPrefix is true: "Registy:HKEY_LOCAL_MACHINE\SOFTWARE\Policies\Microsoft\Windows Defender\Exclusions\Extensions"
function GetPathToRegItem([string]$settingSuffix, [bool]$addRegPrefix = $false)
{
    return (Join-Path (GetPathToKeyDir $local:settingSuffix $local:addRegPrefix) (GetRegValueName $local:settingSuffix))
}

# use GetRegValueType funcation to get the type of the value stored in the registry (of type [Microsoft.Win32.RegistryValueKind]).
function GetRegValueType([string]$nameSuffix)
{
    # if reg value type is different than DWord, return it.
    if(($settingsRegMapping[$settingsNamesMapping.$local:nameSuffix]).Count -gt 2)
    {
        return $settingsRegMapping[$settingsNamesMapping.$local:nameSuffix][([regInfoIndex]::type)]
    }

    return [Microsoft.Win32.RegistryValueKind]::DWord
}

# GetExclusionAddionalSettingName - 
# For exclusions setting in GP hive, additional properties need to 
# be set under "Extensions" key in GP registry hive. This function
# returns the property name according to full extension setting name received
# For invalid setting name, $null is returned
function GetExclusionAddionalSettingName([string] $fullName)
{
    $additionalName = $null
    switch($setting)
    {
        $settingsNamesMapping.ExcludedExtensions { $additionalName = "Exclusions_Extensions"; break}
        $settingsNamesMapping.ExcludedPaths { $additionalName = "Exclusions_Paths"; break}
        $settingsNamesMapping.ExcludedProcesses { $additionalName = "Exclusions_Processes"; break}
        default { LogWarning (-join("AV GetExclusionAddionalSettingName: invalid setting name ",$setting))}
    }

    return $additionalName
}

function RemoveExclusions([string] $setting, [string] $value) {

    $exclusionDirPath = (Join-Path $script:GPRegPath $settingsRegMapping[$local:setting][([regInfoIndex]::dir)])
    $exclusionRegPath = (Join-Path $local:exclusionDirPath $settingsRegMapping[$local:setting][([regInfoIndex]::name)])

    # if no value given (old senseCM version), clear will be based on content in GP regisry (Remove-MpPreference does not accept empty values)
    $exclusionsNamesArr = (Get-Item -Path $local:exclusionRegPath | Select-Object -ExpandProperty Property)

    if([string]::IsNullOrEmpty($local:value)) #if no value is sent in remove we delete all values
    {
        LogInfo (-join("AV: remove ", $local:setting, " with no value, deleting all content: ",($local:exclusionsNamesArr -join('|'))," from Reg:",$local:exclusionRegPath))
        Remove-Item -Path $local:exclusionRegPath
    }
    else 
    {
        #value has content, use it to delete requested value only:
        $exclusionsNamesArr = $local:value.Split("|") | Where-Object {$_} #removing empty values

        foreach($exclusionName in $local:exclusionsNamesArr) 
        {
            LogInfo (-join("AV: Removing exclusion Name: ",$local:exclusionName, " from path ",($local:exclusionRegPath -join('|') )))
            Remove-ItemProperty -Path $local:exclusionRegPath -Name $local:exclusionName -ErrorAction SilentlyContinue #if item doesn't exits, continue.
        }
    }

    #if no values left, remove additional properties under "exclusions" key:
    if((Get-ItemProperty -Path $local:exclusionRegPath).Count -eq 0)
    {
        LogInfo (-join("AV: no items left under ", $local:exclusionRegPath, ". removing ", (GetExclusionAddionalSettingName $local:setting))) 
        Remove-ItemProperty -Path $local:exclusionDirPath -Name (GetExclusionAddionalSettingName $local:setting)
    }

    # as a workaround for issue #39416943 of Defender - trigger an engine config refresh
    TriggerGPRefresh $setting
}

# as a workaround for issue #39416943 of Defender, need to set and remove a dummy value for AV exclusions by Cmdlet to trigger the GP refresh
function TriggerGPRefresh([string] $setting)
{
    switch($local:setting)
    {
        $settingsNamesMapping.ExcludedExtensions { 
            Add-MpPreference -ExclusionExtension $script:dummyValue
            Remove-MpPreference -ExclusionExtension $script:dummyValue
            break
        }
        $settingsNamesMapping.ExcludedPaths { 
            Add-MpPreference -ExclusionPath $script:dummyValue
            Remove-MpPreference -ExclusionPath $script:dummyValue
            break
        }
        $settingsNamesMapping.ExcludedProcesses { 
            Add-MpPreference -ExclusionProcess $script:dummyValue
            Remove-MpPreference -ExclusionProcess $script:dummyValue
            break
        }
        default { LogWarning (-join("AV TriggerGPRefresh: invalid setting name ",$local:setting))}
    }
}

#Setting the exclusions settings according to the array that was filled in 'Set' function. 
#Values that were removed from a specific policy will be removed here even though we are not getting a 'remove' command for them (since the setting still exists in the policy).
function SetExclusionsSettings()
{
    foreach($setting in $script:exclusionValuesToBeSet.Keys)
    {
        try
        {
            if($script:exclusionValuesToBeSet[$local:setting].Count -ne 0)
            {
                # check if path exists, if not - create the dir (Extensions/Paths/Processes)
                $exclusionDirPath = (Join-Path $script:GPRegPath $settingsRegMapping.$local:setting[([regInfoIndex]::dir)])
                $exclusionPath = (Join-Path $local:exclusionDirPath $settingsRegMapping.$local:setting[([regInfoIndex]::name)])

                    if(!(Test-Path $local:exclusionPath)) 
                    {
                        LogInfo (-join("AV SetExclusionsSettings: reg key ",$local:exclusionPath, " doesn't exist. creating."))
                        New-Item -Path $exclusionPath -Force | Out-Null
                    } else {
                        #remove all items (we will re-set the ones we need, and must remove the ones that were deleted from the policy but were not Removed)
                        $currentContent = (Get-Item -Path $local:exclusionPath | Select-Object -ExpandProperty Property)
                        LogInfo (-join("AV SetExclusionsSettings: reg key ",$local:exclusionPath, " exists. Clearing previous content: ",($local:currentContent -join('|') )))
                        Clear-Item $local:exclusionPath
                    }

                LogInfo (-join("AV: setting property ", (GetExclusionAddionalSettingName $local:setting), " under ", $local:exclusionDirPath))
                Set-ItemProperty -Path $local:exclusionDirPath -Name (GetExclusionAddionalSettingName $local:setting) -Type ([Microsoft.Win32.RegistryValueKind]::DWord) -Value "1"

                #add all required content
                foreach($exclusionName in $script:exclusionValuesToBeSet[$local:setting]) 
                {
                    #when GP sets exclusions the value is not set. If fail on creation, we will discover and report on VerifyAssignment.
                    New-ItemProperty -Path $local:exclusionPath -Name $local:exclusionName  -ErrorAction SilentlyContinue | Out-Null 
                }

                # as a workaround for issue #39416943 of Defender, need to set and remove a dummy value for AV exclusions by Cmdlet to trigger the GP refresh
                LogDebug (-join("AV: Triggering GP refresh for ", (GetExclusionAddionalSettingName $local:setting)))
                TriggerGPRefresh $local:setting
            }
        }
        catch 
        {
            LogWarning (-join("AV::CompletionActions failure in SetExclusionsSettings for setting:", $local:setting, " HR:", $_.Exception.HResult, ", Message:", $_.Exception.Message))
        }
    }
}

#used to set the actions for detected threats in CompletionActions
function SetActionForThreats()
{
    LogInfo "AV: Setting actions for detected threats."
         
    $nameSuffix = $settingsNamesMapping.ThreatSeverityDefaultAction

    $threatsDirPath = GetPathToKeyDir $local:nameSuffix $true

    foreach($threatLevel in $script:requestedThreatActions.keys)
    {      
        try
        {
            if($script:conflictUtilsModule."IsSubSettingOnlyFromGroupInConflict"($local:threatLevel, $script:requestedThreatActions, $script:conflictingThreatGroups))
            {
                continue #we don't want to set the threat level action if we got a confilct. 
            }
            $groups = $script:conflictUtilsModule."GetGroupFromSettings"($local:threatLevel, $script:requestedThreatActions)

            $settingErrorInformation = $script:conflictUtilsModule."GetFirstErrorInfoByGroup"($local:groups[0], $script:requestedThreatActions)
            if($settingErrorInformation.Count -gt 0) #check if there is an error logged in $threatActionFirstErrorInfo struct
            {
                continue
            }

            if(!(Test-Path $local:threatsDirPath)) 
            {
                LogInfo (-join("AV: reg key doesn't exist. creating: ", $local:threatsDirPath))
                New-Item -Path $local:threatsDirPath -Force | Out-Null
            }
            $action = $script:requestedThreatActions[$local:threatLevel][([threatInfoIndex]::action)]
            Set-ItemProperty -Path $local:threatsDirPath -Name $local:threatLevel -Value $local:action -Type (GetRegValueType $nameSuffix) -ErrorAction Stop 
            switch($threatLevel)
            {
                1 { Set-MpPreference -LowThreatDefaultAction $local:action; break}
                2 { Set-MpPreference -ModerateThreatDefaultAction $local:action; break}
                4 { Set-MpPreference -HighThreatDefaultAction $local:action; break}
                5 { Set-MpPreference -SevereThreatDefaultAction $local:action; break}
            }
            #If fail on creation, we will discover and report on VerifyAssignment.
        }
        catch {
            LogWarning (-join("AV::CompletionActions failure for SetActionForThreats" ,". HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
            #if an unexpected exception happened, save it in the general error info
            if($threatActionFirstErrorInfo.count -eq 0) #if not 0, we don't want to override the previous error
            {
                #no errors yet, add current error
                $threatActionFirstErrorInfo.Add($_.Exception.HResult,$_.Exception.Message)
            }
        }
    }
}

function RemoveThreatActions()
{
    $nameSuffix = $settingsNamesMapping.ThreatSeverityDefaultAction
    $threatsDirPath = GetPathToKeyDir $local:nameSuffix $true

    try {
        $pathExist = Test-Path $local:threatsDirPath
        if(!$pathExist)
        {
            LogInfo (-join("AV:RemoveThreatActions path: ",$local:threatsDirPath, " does not exist. Nothing to clear from registry."))
        }
        else
        {
            $currentContentObj = (Get-ItemProperty -Path $local:threatsDirPath)

            #for info: print current registry keys in the GP hive from the GP reg key object:
            $currentContentForPrint = $currentContentObj | Select-Object -Property @($script:threatLevelsPrefNamesMapping.Keys)
            LogInfo (-join("AV:RemoveThreatActions from ",$local:threatsDirPath,". Clearing previous content(",$threatsStr,"): ", ($local:currentContentForPrint -join "|") ))
            
            Clear-Item $local:threatsDirPath -ErrorAction Stop
        }

        ClearThreatActionsAccordingToExistingGPValues $local:currentContentObj

    } catch {
        LogWarning (-join("AV:RemoveThreatActions failed to clear values from " , $local:threatsDirPath, ". HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
        $errorCode.Value = $_.Exception.HResult
        $errorString.Value = $_.Exception.Message
    }
}

function ClearThreatActionsAccordingToExistingGPValues([PSObject]$currentGPValue)
{
    foreach($threatLevel in $script:threatLevelsPrefNamesMapping.Keys)
    {
       if(($null -ne $local:currentGPValue) -and ($null -ne $local:currentGPValue.$threatLevel))
       {
            #value existed in GP hive - we need to remove it to handle implicit removals with cmdlets
            # values in the switch are based on: $script:threatLevelsPrefNamesMapping
            switch($local:threatLevel)
            {
                "1" {Remove-MpPreference -LowThreatDefaultAction; break;}
                "2" {Remove-MpPreference -ModerateThreatDefaultAction; break;}
                "4" {Remove-MpPreference -HighThreatDefaultAction; break;}
                "5" {Remove-MpPreference -SevereThreatDefaultAction; break;}
                default {        
                    LogWarning (-join("AV:ClearThreatActionsAccordingToExistingValues invalid threat level: " , $local:threatLevel, "."))
                    }
            }
       }
    }
}

function ResetPrefValue ([string] $settingNameSuffix)
{
    switch ($settingsNamesMapping.$local:settingNameSuffix)
        {
            $settingsNamesMapping.CheckForSignaturesBeforeRunningScan { Remove-MpPreference -CheckForSignaturesBeforeRunningScan; break }
            $settingsNamesMapping.CloudBlockLevel { Remove-MpPreference -CloudBlockLevel; break }
            $settingsNamesMapping.CloudExtendedTimeout {  Remove-MpPreference -CloudExtendedTimeout; break }
            $settingsNamesMapping.AllowArchiveScanning {  Remove-MpPreference -DisableArchiveScanning; break }
            $settingsNamesMapping.AllowBehaviorMonitoring {  Remove-MpPreference -DisableBehaviorMonitoring; break }
            $settingsNamesMapping.DisableCatchupFullScan {  Remove-MpPreference -DisableCatchupFullScan; break }
            $settingsNamesMapping.DisableCatchupQuickScan {  Remove-MpPreference -DisableCatchupQuickScan; break }
            $settingsNamesMapping.AllowEmailScanning {  Remove-MpPreference -DisableEmailScanning; break }
            $settingsNamesMapping.AllowIOAVProtection {  Remove-MpPreference -DisableIOAVProtection; break }
            $settingsNamesMapping.AllowRealtimeMonitoring {  Remove-MpPreference -DisableRealtimeMonitoring; break }
            $settingsNamesMapping.AllowFullScanRemovableDriveScanning {  Remove-MpPreference -DisableRemovableDriveScanning; break }
            $settingsNamesMapping.AllowFullScanOnMappedNetworkDrives {  Remove-MpPreference -DisableScanningMappedNetworkDrivesForFullScan; break }
            $settingsNamesMapping.AllowScanningNetworkFiles {  Remove-MpPreference -DisableScanningNetworkFiles; break }
            $settingsNamesMapping.AllowScriptScanning {  Remove-MpPreference -DisableScriptScanning; break }
            $settingsNamesMapping.EnableLowCpuPriority {  Remove-MpPreference -EnableLowCpuPriority; break }
            $settingsNamesMapping.EnableNetworkProtection { Remove-MpPreference -EnableNetworkProtection; break } #not removing additional related settings 
            $settingsNamesMapping.AllowCloudProtection { Remove-MpPreference -MAPSReporting; break }
            $settingsNamesMapping.PUAProtection { Remove-MpPreference -PUAProtection; break } 
            $settingsNamesMapping.DaysToRetainCleanedMalware {  Remove-MpPreference -QuarantinePurgeItemsAfterDelay; break }
            $settingsNamesMapping.RealTimeScanDirection { Remove-MpPreference -RealTimeScanDirection; break }
            $settingsNamesMapping.AvgCPULoadFactor {  Remove-MpPreference -ScanAvgCPULoadFactor; break }
            $settingsNamesMapping.ScanParameter { Remove-MpPreference -ScanParameters; break }
            $settingsNamesMapping.ScheduleScanDay { Remove-MpPreference -ScanScheduleDay; break }
            $settingsNamesMapping.ScheduleQuickScanTime {  Remove-MpPreference -ScanScheduleQuickScanTime; break }
            $settingsNamesMapping.ScheduleScanTime {  Remove-MpPreference -ScanScheduleTime; break }
            $settingsNamesMapping.SignatureUpdateFileSharesSources {  Remove-MpPreference -SignatureDefinitionUpdateFileSharesSources; break }
            $settingsNamesMapping.SignatureUpdateFallbackOrder {  Remove-MpPreference -SignatureFallbackOrder; break }
            $settingsNamesMapping.SignatureUpdateInterval {  Remove-MpPreference -SignatureUpdateInterval; break }
            $settingsNamesMapping.SubmitSamplesConsent { Remove-MpPreference -SubmitSamplesConsent; break }
            $settingsNamesMapping.AllowUserUIAccess {  Remove-MpPreference -UILockdown; break }
            default 
            {
                LogInfo (-join("AV: ResetPrefValue for ", $local:settingNameSuffix, " - no cmdlet available")) 
            }
        }
}

function SetPrefValue([string] $settingNameSuffix, [string] $newValue)
{
    if ($booleanValues.Contains($settingsNamesMapping.$local:settingNameSuffix))
    {
        $boolVal = GetBoolFromValue $local:newValue
    }

    switch ($settingsNamesMapping.$local:settingNameSuffix)
        {
            # we intentionally don't use invoke expression to avoid code injection in case the policy is malicious
            $settingsNamesMapping.CheckForSignaturesBeforeRunningScan { Set-MpPreference -CheckForSignaturesBeforeRunningScan $local:boolVal; break }
            $settingsNamesMapping.CloudBlockLevel {  Set-MpPreference -CloudBlockLevel $local:newValue; break }
            $settingsNamesMapping.CloudExtendedTimeout {  Set-MpPreference -CloudExtendedTimeout ([System.Convert]::ToInt32($local:newValue)); break }
            $settingsNamesMapping.AllowArchiveScanning {  Set-MpPreference -DisableArchiveScanning $local:boolVal; break }
            $settingsNamesMapping.AllowBehaviorMonitoring {  Set-MpPreference -DisableBehaviorMonitoring $local:boolVal; break }
            $settingsNamesMapping.DisableCatchupFullScan {  Set-MpPreference -DisableCatchupFullScan $local:boolVal; break  }
            $settingsNamesMapping.DisableCatchupQuickScan { Set-MpPreference -DisableCatchupQuickScan $local:boolVal; break  }
            $settingsNamesMapping.AllowEmailScanning {  Set-MpPreference -DisableEmailScanning $local:boolVal; break }
            $settingsNamesMapping.AllowIOAVProtection {  Set-MpPreference -DisableIOAVProtection $local:boolVal; break }
            $settingsNamesMapping.AllowRealtimeMonitoring {  Set-MpPreference -DisableRealtimeMonitoring $local:boolVal; break }
            $settingsNamesMapping.AllowFullScanRemovableDriveScanning {  Set-MpPreference -DisableRemovableDriveScanning $local:boolVal; break }
            $settingsNamesMapping.AllowFullScanOnMappedNetworkDrives {  Set-MpPreference -DisableScanningMappedNetworkDrivesForFullScan $local:boolVal; break }
            $settingsNamesMapping.AllowScanningNetworkFiles {  Set-MpPreference -DisableScanningNetworkFiles $local:boolVal; break }
            $settingsNamesMapping.AllowScriptScanning {  Set-MpPreference -DisableScriptScanning $local:boolVal; break }
            $settingsNamesMapping.EnableLowCpuPriority {  Set-MpPreference -EnableLowCpuPriority $local:boolVal; break }
            $settingsNamesMapping.EnableNetworkProtection { 
                if(IsDLServer)
                {
                    # these settings are required for NW Protection on DL servers
                    SetupNWProtectionForDLServer
                }

                Set-MpPreference -EnableNetworkProtection $local:newValue 
                break 
            }
            $settingsNamesMapping.AllowCloudProtection {  Set-MpPreference -MAPSReporting $local:newValue; break } 
            $settingsNamesMapping.PUAProtection {  Set-MpPreference -PUAProtection $local:newValue; break }
            $settingsNamesMapping.DaysToRetainCleanedMalware {  Set-MpPreference -QuarantinePurgeItemsAfterDelay $local:newValue; break }
            $settingsNamesMapping.RealTimeScanDirection {  Set-MpPreference -RealTimeScanDirection $local:newValue; break }
            $settingsNamesMapping.AvgCPULoadFactor {  Set-MpPreference -ScanAvgCPULoadFactor ([System.Convert]::ToByte($local:newValue)); break }
            $settingsNamesMapping.ScanParameter {  Set-MpPreference -ScanParameters $local:newValue; break }
            $settingsNamesMapping.ScheduleScanDay {  Set-MpPreference -ScanScheduleDay $local:newValue; break }
            $settingsNamesMapping.ScheduleQuickScanTime {  Set-MpPreference -ScanScheduleQuickScanTime ("{0:hh\:mm\:ss}" -f [timespan]::fromminutes($local:newValue)); break }
            $settingsNamesMapping.ScheduleScanTime {  Set-MpPreference -ScanScheduleTime ("{0:hh\:mm\:ss}" -f [timespan]::fromminutes($local:newValue)); break }
            $settingsNamesMapping.SignatureUpdateFileSharesSources {  Set-MpPreference -SignatureDefinitionUpdateFileSharesSources $local:newValue; break }
            $settingsNamesMapping.SignatureUpdateFallbackOrder {  Set-MpPreference -SignatureFallbackOrder $local:newValue; break }
            $settingsNamesMapping.SignatureUpdateInterval {  Set-MpPreference -SignatureUpdateInterval $local:newValue; break }
            $settingsNamesMapping.SubmitSamplesConsent {  Set-MpPreference -SubmitSamplesConsent $local:newValue; break }
            $settingsNamesMapping.AllowUserUIAccess {  Set-MpPreference -UILockdown $local:boolVal; break }
            default 
            {
                LogInfo (-join("AV: SetPrefValue for ", $local:settingNameSuffix, " - no cmdlet available")) 
            }
        }
}

# Bool value can either be represented by "true"\"false" string or by number
function GetBoolFromValue([string] $value)
{
    if([string]::IsNullOrEmpty($local:value))
    {
        return $null
    }
    if ($local:value -ieq "true")
    {
        return $true
    }
    elseif ($local:value -ieq "false") 
    {
        return $false
    }
    elseif ([System.Convert]::ToInt32($local:value) -ne 0) 
    {
        return $true
    }
    else 
    {
        return $false
    }
}

function GetValidPrefix([string] $fullName)
{
    $validPrefix = $null
    foreach($prefix in $script:settingNamePrefixes)
    {
        if($local:fullName.StartsWith($prefix))
        {
            $validPrefix = $prefix
            break
        }
    }
    return $local:validPrefix
}

# GetNameSuffix returns the "clean" setting name,
# without the prefix.
# IMPORTANT - this function assumes the setting name was validated by IsSupportedParam(), 
# meaning - it has a valid prefix.
function GetNameSuffix([string] $fullName)
{
    if([string]::IsNullOrEmpty($local:fullName))
    {
        return $null
    }

    $settingPrefix = GetValidPrefix -fullName $local:fullName
    return ($local:fullName.Replace($local:settingPrefix, ""))
}

function GetNameSuffixAndSettingPrefix([string] $name, [ref] $nameSuffix, [ref] $settingPrefix)
{
    $nameSuffix.Value = GetNameSuffix -fullName $name 
    $settingPrefix.Value = GetSettingPrefix($nameSuffix.Value)
    if ($settingPrefix.Value -ne "")
    {
        $nameSuffix.Value = $nameSuffix.Value.Replace($settingPrefix.Value + ".", "")
    }
}

function GetSettingCurrentValueFromMpPref([string] $nameSuffix, [ref] $errorCode, [ref] $errorString, [ref] $hresult)
{
    $hresult.Value = [HResult]::S_OK
    
    $local:mpPref = Get-MpPreference
    if($null -eq $local:mpPref)
    {
        LogWarning "AV::GetSettingCurrentValueFromMpPref failure. Failed to get Mp-Preference."
        $hresult.value = [HResult]::E_FAIL
        return $null
    }

    switch($settingsNamesMapping.$local:nameSuffix)
    {
        {$_ -in $script:exclusionValuesToBeSet.Keys} {  $local:settingCurrentValue = GetExclusionsCurrentValueFromMpPref $local:nameSuffix $local:mpPref $errorCode $errorString $hresult  }
        {$_ -eq $settingsNamesMapping.ThreatSeverityDefaultAction} {$local:settingCurrentValue = GetThreatActionsCurrentValueFromMpPref $local:mpPref}
        Default {   $local:settingCurrentValue = $local:mpPref.$_ }

    }
    
    if ($local:settingCurrentValue.Length -eq 0)
    {
        LogWarning "AV:GetSettingCurrentValueFromMpPref - current value is empty." #current value should not be empty here since we call this in VerAssign after Set.
        $local:settingCurrentValue = $null
    }

    if ($settingsNamesMapping.$local:nameSuffix -eq $settingsNamesMapping.EnableNetworkProtection)
    {
        if(IsDLServer)
        {
            #for DL server, if additional setting were not configured, EnableNetworkProtection will not work:
            if(!($local:mpPref.AllowNetworkProtectionOnWinServer) -or !($local:mpPref.AllowNetworkProtectionDownLevel) -or !($local:mpPref.AllowDatagramProcessingOnWinServer))
            {
                $errorCode.Value = [HResult]::E_FAIL.value__
                $msg = -join("AV:GetSettingCurrentValueFromMpPref for EnableNetworkProtection, additional settings not set: AllowNetworkProtectionOnWinServer:",$local:mpPref.AllowNetworkProtectionOnWinServer," AllowNetworkProtectionDownLevel:",$local:mpPref.AllowNetworkProtectionDownLevel," AllowDatagramProcessingOnWinServer:",$local:mpPref.AllowDatagramProcessingOnWinServer)
                $errorString.Value = $msg
                LogWarning $msg
                $local:settingCurrentValue = $null
            }
        }
    }

    if ($settingsNamesMapping.$local:nameSuffix -eq $settingsNamesMapping.AllowCloudProtection)
    {
        # Represent expected & actual values as true for active and false for not-active.
        # Values of 1/2 (Basic/Advanced) are both accepted as 'active'. A value of 0 means cloud protection is 'not-active'
        $isCurrentValueActive = ($local:settingCurrentValue -in $allowCloudProtectionActiveValues) 
        LogInfo (-join("AV:GetSettingCurrentValueFromMpPref for AllowCloudProtection - converting settingCurrentValue value from ",$local:settingCurrentValue," to ",$local:isCurrentValueActive," (true for Active, false for Inactive)"))
        $local:settingCurrentValue = $isCurrentValueActive
    }

    return $local:settingCurrentValue
}

function GetExpectedMergeValue([string] $nameSuffix, [string] $value, [ref] $isExclusions, [ref] $isThreatActions)
{
    if($nameSuffix -eq $settingsNamesMapping.ThreatSeverityDefaultAction)
    {
        $local:expectedValue = $script:conflictUtilsModule."ParseSettingInfo"($value) 
        $isThreatActions.Value = $true
    }
    else { #exclusions
        $local:expectedValue = GetExclusionsSettingValueAsArray $value
        $isExclusions.Value = $true
    }
    return $local:expectedValue
}

function GetValueInRegistryFormat([string] $nameSuffix, [string] $value, [ref] $isExclusions, [ref] $isThreat)
{
    $local:expectedValue = $null
    $isExclusions.Value = $false
    switch ($settingsNamesMapping.$nameSuffix)
    {
        {$_ -in $booleanValues} 
        {
            $expectedValue = $value 
            if ($_ -in $reverseBooleanValues)
            {
                $reverseExpectedValue = 1 - $expectedValue
                LogInfo "AV:GetValueInRegistryFormat for $nameSuffix - toggling expected value from $local:expectedValue to $local:reverseExpectedValue"
                $expectedValue = $reverseExpectedValue #toggling int representing bool (0/1)
            }
            break
        }
        {$_ -in $script:mergeSettingsNamesMapping}
        {
            $local:expectedValue = GetExpectedMergeValue $nameSuffix $value $isExclusions $isThreatAction
            break
        }
        {$settingsNamesMapping.ContainsValue($_)} # all "special" handling of settings was done in previous cases. Now handle all the "regular" settings 
        {
            $local:expectedValue = $value
        }
    }  
    
    return $local:expectedValue
}

# This function takes the expected value we got from MEM in string format and returns the corresponding value they is returned from Get-MpPreference
function GetValueInCmdletFormat([string] $nameSuffix, [string] $value, [ref] $isExclusions, [ref] $isThreatActions, [ref] $errorCode, [ref] $errorString)
{   
    $local:expectedValue = $null
    switch ($settingsNamesMapping.$nameSuffix)
    {
        {$_ -in $booleanValues} 
        {
            $local:expectedValue = GetBoolFromValue $value
            if ($_ -in $reverseBooleanValues)
            {
                $local:expectedValue = !$local:expectedValue
            }
            break
        }
        {$_ -eq $settingsNamesMapping.AvgCPULoadFactor}
        {
            $local:expectedValue = [System.Convert]::ToByte($value)
            break
        }
        {$_ -eq $settingsNamesMapping.AllowOnAccessProtection }
        { 
            $local:expectedValue = 1 - $value #flipping the value for the setting
            LogInfo (-join("AV:GetValueInCmdletFormat for AllowOnAccessProcetion - toggling expected value from ",$value," to ",$local:expectedValue))
            break 
        }
        {$_ -eq $settingsNamesMapping.AllowCloudProtection}
        {
            # Represent expected & actual values as true for active and false for not-active.
            # Values of 1/2 (Basic/Advanced) are both accepted as 'active'. A value of 0 means cloud protection is 'not-active'
            $local:isExpectedValueActive =  ($value -in $allowCloudProtectionActiveValues) 
            LogInfo "AV:GetValueInCmdletFormat for AllowCloudProtection - converting expected value from $value to $local:isExpectedValueActive (true for Active, false for Inactive)"
            
            $local:expectedValue = $local:isExpectedValueActive
            break
        }
        {$_ -in $script:mergeSettingsNamesMapping}
        {
            $local:expectedValue = GetExpectedMergeValue $nameSuffix $value $isExclusions $isThreatActions
            break
        }
        {($_ -eq $settingsNamesMapping.ScheduleQuickScanTime) -or ($_ -eq $settingsNamesMapping.ScheduleScanTime) }
        {
            $local:expectedValue = "{0:hh\:mm\:ss}" -f [timespan]::fromminutes($value)
            break
        }
        {$settingsNamesMapping.ContainsKey($nameSuffix)} 
        {
            $local:expectedValue = $value
            break
        }
        Default 
        {
            $errorCode.Value = [HResult]::E_NOINTERFACE.value__
            $errorString.Value = (WriteErrorSettingIsNotSupported -funcName "VerifyAssignment" -name $nameSuffix)
            
            $local:expectedValue = $null
        }
    }

    return $local:expectedValue
}

function IsTamperProtectionActiveOnSetting([string] $nameSuffix)
{
    #TP1 - marker for tests - don't remove!

    return ($settingsNamesMapping.$local:nameSuffix -in $script:settingBlockedByTamperProtection) -and (IsTamperProtectionActive)
}

function GetSettingCurrentValueFromGPRegistry([string] $nameSuffix, [bool] $isExclusions)
{
    $local:settingCurrentValue = $null
    if ($isExclusions)
    {
        $local:settingCurrentValue = GetCurrentValueOfRegistryDir $local:nameSuffix
    } 
    else 
    {
        $local:settingCurrentValue = GetSingleCurrentValueFromRegistry $local:nameSuffix
    }

    if(($local:settingCurrentValue.Length -eq 0))
    {
        $emptyRegStr = "GP(" + (GetPathToKeyDir $local:nameSuffix $true)  + ")"
        LogWarning "AV:GetSettingCurrentValueFromGPRegistry - cannot get current value from GP Reg hive: $emptyRegStr"
    }

    return $local:settingCurrentValue
}

function GetEnforceSucceeded([string[]] $settingCurrentValue, [string[]] $expectedValue)
{
    $local:enforceSucceeded = $true 
   
    foreach($singleValue in $expectedValue)
    {
        if(!($singleValue -in $settingCurrentValue))
        {
            $local:enforceSucceeded = $false
            break
        }
    }

    return $local:enforceSucceeded
}

function GetExclusionsSettingValueAsArray([string] $exclusionsSettingValue)
{
    $local:exclusionsSettingValueAsArray = $exclusionsSettingValue.Split("|")
    $local:exclusionsSettingValueAsArray = $local:exclusionsSettingValueAsArray | ForEach-Object { $_.Trim() | Where-Object {$_}} #removing empty values and trimming spaces
    return $local:exclusionsSettingValueAsArray
}

function GetExclusionsCurrentValueFromMpPref([string] $nameSuffix, [ciminstance] $mpPref, [ref] $errorCode, [ref] $errorString, [ref] $hresult)
{
    $settingCurrentValue = $local:mpPref.($settingsNamesMapping.$local:nameSuffix)
    $defenderRegValue = Get-ItemProperty $script:GPRegPath
    $HideExclusionValue = $local:defenderRegValue.HideExclusionsFromLocalAdmins
    if($local:HideExclusionValue -eq 1)
    {
        $msg = -join("Cannot get value of ", $local:nameSuffix, " because HideExclusionsFromLocalAdmins is true")
        LogWarning $msg
        $errorCode.Value = [HResult]::E_ACCESSDENIED.value__
        $errorString.Value = $msg
        $hresult.Value = [HResult]::E_FAIL
        $local:settingCurrentValue = $null
    }
    $settingCurrentValue = $settingCurrentValue.ForEach({ $_.Trim() })#in DL servers Get-MpPreference output has spaces in the values, so we need to remove them for validateValuesAssignment
    return $local:settingCurrentValue
} 

function GetThreatActionsCurrentValueFromMpPref([ciminstance] $mpPref)
{
    $valuesMap = @{}
    foreach($level in $script:threatLevelsPrefNamesMapping.Keys)
    {
        $local:valuesMap.Add([string]$level, $mpPref.($script:threatLevelsPrefNamesMapping[$level]))
    }
    return $local:valuesMap
}

#ValidateThreatsAssignment- Validates that all expected values for a specific policy were Set to settingCurrentValue - map of (level=action) we get from Get-MpPreference.
function ValidateThreatsAssignment([hashtable] $settingCurrentValue, [hashtable] $expectedValue, [ref] $errorCode, [ref] $errorString)
{
    $expectedLevels = @($expectedValue.Keys)
    $groups = $script:conflictUtilsModule."GetGroupFromSettings"($local:expectedLevels, $script:requestedThreatActions)

    if($script:conflictingThreatGroups -contains $local:groups[0]) #check if this group is in a conflict. If more than 1 group is returned, means groups are identical, so checking first. Setting is added to 'conflictingThreatGroups' in 'StoreSettingToRequested' in the ConflictUtils module.
    {
        $errorCode.Value = [HResult]::E_INTERNAL_CONFLICT.value__
        $errorString.Value = "conflict detected for requested value based on group: " + $local:groups[0]
        return $false
    }

    foreach($level in $local:expectedLevels) #check that values for all expected levels match:
    {   
        $settingErrorInformation = $script:conflictUtilsModule."GetFirstErrorInfoByGroup"($local:groups[0], $script:requestedThreatActions)
        if($local:settingCurrentValue.$local:level -ne $local:expectedValue.$local:level) 
        {
            if($local:settingErrorInformation.Count -gt 0) #check if there is an error logged in $threatActionFirstErrorInfo struct
            {
                $errorCode.Value = $local:settingErrorInformation[([firstErrorIndex]::HR)]
                $errorString.Value = $local:settingErrorInformation[([firstErrorIndex]::errStr)]
            }
            LogWarning (-join("AV: ValidateThreatsAssignment value for level:", $local:level, " is different. Expected:", $local:expectedValue.$local:level, ", current:", $local:settingCurrentValue.$local:level))
            return $false
        }
    }

    return $true
}

function IsThreatLevelsAndActionsValid([hashtable] $threatActionsDict, [ref] $errorMsg)
{
    foreach($threatLevel in $threatActionsDict.keys)
    {
        if($local:threatLevel -notin $script:threatLevelsPrefNamesMapping.Keys)
        {
            $msg = -join("AV::IsThreatLevelsAndActionsValid: Threat Level: ", $local:threatLevel , " is not valid.")
            LogWarning $msg
            $errorMsg.Value = -join("Threat Level: ", $local:threatLevel , " is not valid.")
            $script:requestedThreatActions[$local:threatLevel][[threatInfoIndex]::FirstErrorInfo] = @([HResult]::E_INVALIDARG.value__ ,$msg)
            return $false
        }

        $action = $local:threatActionsDict[$local:threatLevel]
        if($local:action -notin $script:supportedThreatActions)
        {
            $msg = -join("AV::IsThreatLevelsAndActionsValid: Action Level: ", $local:action , " is not valid.")
            LogWarning $msg
            $errorMsg.Value = -join("Action Level: ", $local:action , " is not valid.")
            $script:requestedThreatActions[$local:threatLevel][[threatInfoIndex]::FirstErrorInfo] = @([HResult]::E_INVALIDARG.value__ ,$msg)
            return $false
        }
    }
    return $true
}

function PrepareValueToPrint($value)
{
    if($null -eq $value) 
    {
        return $value
    }
    switch ($value.GetType().Name) { #converting different types to string
        "Object[]" {    #array of items
             return ($value -join('|')) 
        } 
        "Hashtable" {  #Hashtable
            $str = ""
            foreach($key in $value.Keys)
            {
                if($str -ne "")
                {
                    $str += "|"
                }
                $str += $key 
                $str += "=" 
                $str += $value[$key]
            }
            return $str
        }
    }
    return $value
}

#RetrieveTestInfo - returns module info for the TestRunner. Returning here any variable and struct needed for running the tests for this module
function RetrieveTestInfo() {
    return $settingsNamesMapping, $settingsRegMapping, $regKeyFolder, $script:settingNamePrefixes, (-join($script:RegPrefix,$script:GPRegPath)), $script:registryOnlySettings, $script:threatLevelsPrefNamesMapping.keys
}
# SIG # Begin signature block
# MIIn4wYJKoZIhvcNAQcCoIIn1DCCJ9ACAQExDzANBglghkgBZQMEAgEFADB5Bgor
# BgEEAYI3AgEEoGswaTA0BgorBgEEAYI3AgEeMCYCAwEAAAQQH8w7YFlLCE63JNLG
# KX7zUQIBAAIBAAIBAAIBAAIBADAxMA0GCWCGSAFlAwQCAQUABCAxmSx5tMAhwzC0
# Icp6fEm6V7rdnq95rqg7p3GVNIFolKCCDZcwggYVMIID/aADAgECAhMzAAADcaWg
# nFyRHuruAAAAAANxMA0GCSqGSIb3DQEBCwUAMH4xCzAJBgNVBAYTAlVTMRMwEQYD
# VQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNpZ25p
# bmcgUENBIDIwMTEwHhcNMjMwNTExMTg0MTEzWhcNMjQwNTA4MTg0MTEzWjCBlDEL
# MAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1v
# bmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjE+MDwGA1UEAxM1TWlj
# cm9zb2Z0IFdpbmRvd3MgRGVmZW5kZXIgQWR2YW5jZWQgVGhyZWF0IFByb3RlY3Rp
# b24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDQl+fOHFGkNfDHrcv3
# g8shIsUp58WmSbtNm4uu7Sw75Wgsgo5p2yYGGqiRWbu0eIKhAJ1bRp9Qa7vAjMEw
# d0wf9wlLjDNAITzDsokwoCl0rOgWxzYuGSusZq4e/liClIpY5fuxrpsEnsssR1o6
# jITkdJQ6ZkWsTFETxLLOEGxe9BkTkXZaBOHlhTt5QIq6UB6j8zNN/Vo4rFmq751X
# lQLA1jOnNi5NOS3/11TW16ZCo9XKG22zdNsJCUckT/QH2eeQNgpuyfIvP06o+RpP
# Uo7P4xdc4c74S3HKfRiNuqbs3+xvb22GtXIz3cKFDhILpNKN/SC8qLjizq0PxwL1
# lr1ZAgMBAAGjggFzMIIBbzAfBgNVHSUEGDAWBggrBgEFBQcDAwYKKwYBBAGCN0wv
# ATAdBgNVHQ4EFgQUcR5wDY+Mh5A/I7scm54A8ODeSVcwRQYDVR0RBD4wPKQ6MDgx
# HjAcBgNVBAsTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEWMBQGA1UEBRMNNDUxODk0
# KzUwMTAwNTAfBgNVHSMEGDAWgBRIbmTlUAXTgqoXNzcitW2oynUClTBUBgNVHR8E
# TTBLMEmgR6BFhkNodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpb3BzL2NybC9N
# aWNDb2RTaWdQQ0EyMDExXzIwMTEtMDctMDguY3JsMGEGCCsGAQUFBwEBBFUwUzBR
# BggrBgEFBQcwAoZFaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraW9wcy9jZXJ0
# cy9NaWNDb2RTaWdQQ0EyMDExXzIwMTEtMDctMDguY3J0MAwGA1UdEwEB/wQCMAAw
# DQYJKoZIhvcNAQELBQADggIBAKYhzPfo4pWE2MYg2W1VnHJ7NZ/nwEpGYEz7Qhqc
# bPj2sJFFhnUzYHOAyXz4vP0+ERhFw9lrqLvoyj8s6xQU11QPQVItopahH65USGWh
# UcyrXFBVSx7MG3Dwi+NT6rxq9MjY3KLB1hZPF7hgtOjWgtjXgeJOsy3IjPiRkDwX
# 5b5Pg6o8Xpyt84mpvAPsDKO7yJGJq8V528deIYh+0DcELvUZTtl4Kz8ElT4BxF6i
# XxXkuqpCWO/rwEa6HrLNV2udND/AImVio09bbZXOWBcb9uTTNezYX5fZvinIX7hl
# UHHIDaVUqqwTFtac8q7BLE02T0P7NAeBtayujIQFzGNuwLN302UmUWAyc4LynqZa
# h2OmZWcMODUCuHIMP3pk54XjWt0YYEWPFAbA7T7j8wjby/vPXdG93LybID/yjGDb
# U1t5xMoEpN5rvSUM1U42iJ1tF268/nE2/8nKqUpdJXH/MaXxYU3FnbshmIe1xfxF
# NjeMYN4jg+5gjz5SBzHAhT1X9QWXNOhAXxib3wcchlsxFYtSNihHLEJ2GuckRuPi
# 47VWi/3hTTnl88S8r3jsbWLatGbQxcbrsF4sglAvFD9SQSXLGUsE5wz/1p/Gb1Fh
# qfNBc6E4AWVj9+A1KezSnMBAw8gJk3fiRkdxfSfpLOkHvgA7qsOmQ7qwjs05kWZf
# 055CMIIHejCCBWKgAwIBAgIKYQ6Q0gAAAAAAAzANBgkqhkiG9w0BAQsFADCBiDEL
# MAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1v
# bmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEyMDAGA1UEAxMpTWlj
# cm9zb2Z0IFJvb3QgQ2VydGlmaWNhdGUgQXV0aG9yaXR5IDIwMTEwHhcNMTEwNzA4
# MjA1OTA5WhcNMjYwNzA4MjEwOTA5WjB+MQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMSgwJgYDVQQDEx9NaWNyb3NvZnQgQ29kZSBTaWduaW5nIFBD
# QSAyMDExMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEAq/D6chAcLq3Y
# bqqCEE00uvK2WCGfQhsqa+laUKq4BjgaBEm6f8MMHt03a8YS2AvwOMKZBrDIOdUB
# FDFC04kNeWSHfpRgJGyvnkmc6Whe0t+bU7IKLMOv2akrrnoJr9eWWcpgGgXpZnbo
# MlImEi/nqwhQz7NEt13YxC4Ddato88tt8zpcoRb0RrrgOGSsbmQ1eKagYw8t00CT
# +OPeBw3VXHmlSSnnDb6gE3e+lD3v++MrWhAfTVYoonpy4BI6t0le2O3tQ5GD2Xuy
# e4Yb2T6xjF3oiU+EGvKhL1nkkDstrjNYxbc+/jLTswM9sbKvkjh+0p2ALPVOVpEh
# NSXDOW5kf1O6nA+tGSOEy/S6A4aN91/w0FK/jJSHvMAhdCVfGCi2zCcoOCWYOUo2
# z3yxkq4cI6epZuxhH2rhKEmdX4jiJV3TIUs+UsS1Vz8kA/DRelsv1SPjcF0PUUZ3
# s/gA4bysAoJf28AVs70b1FVL5zmhD+kjSbwYuER8ReTBw3J64HLnJN+/RpnF78Ic
# V9uDjexNSTCnq47f7Fufr/zdsGbiwZeBe+3W7UvnSSmnEyimp31ngOaKYnhfsi+E
# 11ecXL93KCjx7W3DKI8sj0A3T8HhhUSJxAlMxdSlQy90lfdu+HggWCwTXWCVmj5P
# M4TasIgX3p5O9JawvEagbJjS4NaIjAsCAwEAAaOCAe0wggHpMBAGCSsGAQQBgjcV
# AQQDAgEAMB0GA1UdDgQWBBRIbmTlUAXTgqoXNzcitW2oynUClTAZBgkrBgEEAYI3
# FAIEDB4KAFMAdQBiAEMAQTALBgNVHQ8EBAMCAYYwDwYDVR0TAQH/BAUwAwEB/zAf
# BgNVHSMEGDAWgBRyLToCMZBDuRQFTuHqp8cx0SOJNDBaBgNVHR8EUzBRME+gTaBL
# hklodHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNS
# b29DZXJBdXQyMDExXzIwMTFfMDNfMjIuY3JsMF4GCCsGAQUFBwEBBFIwUDBOBggr
# BgEFBQcwAoZCaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNS
# b29DZXJBdXQyMDExXzIwMTFfMDNfMjIuY3J0MIGfBgNVHSAEgZcwgZQwgZEGCSsG
# AQQBgjcuAzCBgzA/BggrBgEFBQcCARYzaHR0cDovL3d3dy5taWNyb3NvZnQuY29t
# L3BraW9wcy9kb2NzL3ByaW1hcnljcHMuaHRtMEAGCCsGAQUFBwICMDQeMiAdAEwA
# ZQBnAGEAbABfAHAAbwBsAGkAYwB5AF8AcwB0AGEAdABlAG0AZQBuAHQALiAdMA0G
# CSqGSIb3DQEBCwUAA4ICAQBn8oalmOBUeRou09h0ZyKbC5YR4WOSmUKWfdJ5DJDB
# ZV8uLD74w3LRbYP+vj/oCso7v0epo/Np22O/IjWll11lhJB9i0ZQVdgMknzSGksc
# 8zxCi1LQsP1r4z4HLimb5j0bpdS1HXeUOeLpZMlEPXh6I/MTfaaQdION9MsmAkYq
# wooQu6SpBQyb7Wj6aC6VoCo/KmtYSWMfCWluWpiW5IP0wI/zRive/DvQvTXvbiWu
# 5a8n7dDd8w6vmSiXmE0OPQvyCInWH8MyGOLwxS3OW560STkKxgrCxq2u5bLZ2xWI
# UUVYODJxJxp/sfQn+N4sOiBpmLJZiWhub6e3dMNABQamASooPoI/E01mC8CzTfXh
# j38cbxV9Rad25UAqZaPDXVJihsMdYzaXht/a8/jyFqGaJ+HNpZfQ7l1jQeNbB5yH
# PgZ3BtEGsXUfFL5hYbXw3MYbBL7fQccOKO7eZS/sl/ahXJbYANahRr1Z85elCUtI
# EJmAH9AAKcWxm6U/RXceNcbSoqKfenoi+kiVH6v7RyOA9Z74v2u3S5fi63V4Guzq
# N5l5GEv/1rMjaHXmr/r8i+sLgOppO6/8MO0ETI7f33VtY5E90Z1WTk+/gFcioXgR
# MiF670EKsT/7qMykXcGhiJtXcVZOSEXAQsmbdlsKgEhr/Xmfwb1tbWrJUnMTDXpQ
# zTGCGaIwghmeAgEBMIGVMH4xCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5n
# dG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9y
# YXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNpZ25pbmcgUENBIDIwMTEC
# EzMAAANxpaCcXJEe6u4AAAAAA3EwDQYJYIZIAWUDBAIBBQCgga4wGQYJKoZIhvcN
# AQkDMQwGCisGAQQBgjcCAQQwHAYKKwYBBAGCNwIBCzEOMAwGCisGAQQBgjcCARUw
# LwYJKoZIhvcNAQkEMSIEIOnc3eAyC6bxzaCSvNBuEPguFdOlHj9d5ujKowIClY81
# MEIGCisGAQQBgjcCAQwxNDAyoBSAEgBNAGkAYwByAG8AcwBvAGYAdKEagBhodHRw
# Oi8vd3d3Lm1pY3Jvc29mdC5jb20wDQYJKoZIhvcNAQEBBQAEggEAfSxlFE/yS529
# DWfKOZk7I69Eyagal+D/1Qom5ANP0XLlR9uxQ/zXyVoWCJGwF+kS9IUZmkTDZWAb
# +0CYJjh6JJe0olxKZ4BiidWoSt+dVFHN/tKVHdX+DCrufN2MLDZNZM3fYemPhW89
# TYFW1NL9K9vRmRUpZPT1Vd2KvjmYfbR1/jd75ZDEsib97QQzghWy9mJiea8NAL4V
# Jv77k7Y9otA0v6T3S3pnHglbI7sH2ReH9knt1lzKqhdnuIgUi3xU6zVGTOnes/k+
# MjnVGBddPPQB6vc7Ugz6EeevkkXYsY8lWxyJ0RN53nEy+U0JLPNHnKkovR0eo4w7
# +BogYNUEVaGCFywwghcoBgorBgEEAYI3AwMBMYIXGDCCFxQGCSqGSIb3DQEHAqCC
# FwUwghcBAgEDMQ8wDQYJYIZIAWUDBAIBBQAwggFZBgsqhkiG9w0BCRABBKCCAUgE
# ggFEMIIBQAIBAQYKKwYBBAGEWQoDATAxMA0GCWCGSAFlAwQCAQUABCAONjZ1KD8a
# /KyIgiFywi9Xxwr7YiXf3BmpGCLwiUL+mAIGZJNEiXvHGBMyMDIzMDcxMTE3NDcw
# Mi45ODZaMASAAgH0oIHYpIHVMIHSMQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2Fz
# aGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENv
# cnBvcmF0aW9uMS0wKwYDVQQLEyRNaWNyb3NvZnQgSXJlbGFuZCBPcGVyYXRpb25z
# IExpbWl0ZWQxJjAkBgNVBAsTHVRoYWxlcyBUU1MgRVNOOkZDNDEtNEJENC1EMjIw
# MSUwIwYDVQQDExxNaWNyb3NvZnQgVGltZS1TdGFtcCBTZXJ2aWNloIIRezCCBycw
# ggUPoAMCAQICEzMAAAG59gANZVRPvAMAAQAAAbkwDQYJKoZIhvcNAQELBQAwfDEL
# MAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1v
# bmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEmMCQGA1UEAxMdTWlj
# cm9zb2Z0IFRpbWUtU3RhbXAgUENBIDIwMTAwHhcNMjIwOTIwMjAyMjE3WhcNMjMx
# MjE0MjAyMjE3WjCB0jELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24x
# EDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlv
# bjEtMCsGA1UECxMkTWljcm9zb2Z0IElyZWxhbmQgT3BlcmF0aW9ucyBMaW1pdGVk
# MSYwJAYDVQQLEx1UaGFsZXMgVFNTIEVTTjpGQzQxLTRCRDQtRDIyMDElMCMGA1UE
# AxMcTWljcm9zb2Z0IFRpbWUtU3RhbXAgU2VydmljZTCCAiIwDQYJKoZIhvcNAQEB
# BQADggIPADCCAgoCggIBAONJPslh9RbHyQECbUIINxMF5uQkyN07VIShITXubLpW
# nANgBCLvCcJl7o/2HHORnsRcmSINJ/qclAmLIrOjnYnrbocAnixiMEXC+a1sZ84q
# xYWtEVY7VYw0LCczY+86U/8shgxqsaezKpWriPOcpV1Sh8SsOxf30yO7jvld/IBA
# 3T6lHM2pT/HRjWk/r9uyx0Q4atx0mkLVYS9y55/oTlKLE00h792S+maadAdy3VgT
# weiwoEOXD785wv3h+fwH/wTQtC9lhAxhMO4p+OP9888Wxkbl6BqRWXud54RTzqp2
# Vr+yen1Q1A6umyMB7Xq0snIYG5B1Acc4UgJlPQ/ZiMkqgxQNFCWQvz0G9oLgSPD8
# Ky0AkX22PcDOboPuNT4RceWPX0UVZUsX9IUgs7QF41HiQSwEeOOHGyrfQdmSslAT
# rbmH/18M5QrsTM5JINjct9G42xqN8VF9Z8WOiGMjNbvlpcEmmysYl5QyhrEDoFnQ
# TU7bFrD3JX0fIfu1sbLWeBqXwbp4Z8yACTtphK2VbzOvi4vc0RCmRNzvYQQ2PjZ7
# NaTXE4Gu3vggAJ+rtzUTAfJotvOSqcMgNwLZa1Y+ET/lb0VyjrYwFuHtg0QWyQjP
# 5350LTpv086pyVUh4A3w/Os5hTGFZgFe5bCyMnpY09M0yPdHaQ/56oYUsSIcyKyV
# AgMBAAGjggFJMIIBRTAdBgNVHQ4EFgQUt7A4cdtYQ5oJjE1ZqrSonp41RFIwHwYD
# VR0jBBgwFoAUn6cVXQBeYl2D9OXSZacbUzUZ6XIwXwYDVR0fBFgwVjBUoFKgUIZO
# aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraW9wcy9jcmwvTWljcm9zb2Z0JTIw
# VGltZS1TdGFtcCUyMFBDQSUyMDIwMTAoMSkuY3JsMGwGCCsGAQUFBwEBBGAwXjBc
# BggrBgEFBQcwAoZQaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraW9wcy9jZXJ0
# cy9NaWNyb3NvZnQlMjBUaW1lLVN0YW1wJTIwUENBJTIwMjAxMCgxKS5jcnQwDAYD
# VR0TAQH/BAIwADAWBgNVHSUBAf8EDDAKBggrBgEFBQcDCDAOBgNVHQ8BAf8EBAMC
# B4AwDQYJKoZIhvcNAQELBQADggIBAM3cZ7NFUHRMsLKzjl7rJPIkv7oJ+s9kkut0
# hZif9WSt60SzYGULp1zmdPqc+w8eHTkhqX0GKCp2TTqSzBXBhwHOm8+p6hUxNlDe
# wGMZUos952aTXblAT3OKBnfVBLQyUavrSjuJGZAW30cNY3rjVDUlGD+VygQHySaD
# aviJQbK6/6fQvUUFoqIk3ldGfjnAtnebsVlqh6WWamVc5AZdpWR1jSzN/oxKYqc1
# BG4SxxlPtcfrAdBz/cU4bxVXqAAf02NZscvJNpRnOALf5kVo2HupJXCsk9TzP5PN
# W2sTS3TmwhIQmPxr0E0UqOojUrBJUOhbITAxcnSa/IMluL1HXRtLQZI+xs2eRtuP
# OUsKUW71/1YeqsYCLHLvu82ceDVQQvP7GHEEkp2kEjiofbjYErBo2iCEaxxeX4Z9
# HvAgA4MsQkbn6e4EFQf13sP+Kn3XgMIvJbqLJeFcQja+SUeOXu5cfkxe0GzTNojd
# yIwzaHlhOflVRZNrxee3B+yZwd3JHDIvv71uSI/SIzzt9cU2GyHQVqxBSrRtKW6W
# 8Vw7zpVvoVsIv3ljxg+7NiGSlXX1s7zbBNDMUj9OnzOlHK/3mrOU8YEuRf6RwakW
# 5UCeGamy5MiKu2YuyKiGBCv4OGhPstNe7ALkEOh8BX12t4ntuYu+gw9L6yCPY0jW
# YaQtzAP9MIIHcTCCBVmgAwIBAgITMwAAABXF52ueAptJmQAAAAAAFTANBgkqhkiG
# 9w0BAQsFADCBiDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAO
# BgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEy
# MDAGA1UEAxMpTWljcm9zb2Z0IFJvb3QgQ2VydGlmaWNhdGUgQXV0aG9yaXR5IDIw
# MTAwHhcNMjEwOTMwMTgyMjI1WhcNMzAwOTMwMTgzMjI1WjB8MQswCQYDVQQGEwJV
# UzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UE
# ChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSYwJAYDVQQDEx1NaWNyb3NvZnQgVGlt
# ZS1TdGFtcCBQQ0EgMjAxMDCCAiIwDQYJKoZIhvcNAQEBBQADggIPADCCAgoCggIB
# AOThpkzntHIhC3miy9ckeb0O1YLT/e6cBwfSqWxOdcjKNVf2AX9sSuDivbk+F2Az
# /1xPx2b3lVNxWuJ+Slr+uDZnhUYjDLWNE893MsAQGOhgfWpSg0S3po5GawcU88V2
# 9YZQ3MFEyHFcUTE3oAo4bo3t1w/YJlN8OWECesSq/XJprx2rrPY2vjUmZNqYO7oa
# ezOtgFt+jBAcnVL+tuhiJdxqD89d9P6OU8/W7IVWTe/dvI2k45GPsjksUZzpcGkN
# yjYtcI4xyDUoveO0hyTD4MmPfrVUj9z6BVWYbWg7mka97aSueik3rMvrg0XnRm7K
# MtXAhjBcTyziYrLNueKNiOSWrAFKu75xqRdbZ2De+JKRHh09/SDPc31BmkZ1zcRf
# NN0Sidb9pSB9fvzZnkXftnIv231fgLrbqn427DZM9ituqBJR6L8FA6PRc6ZNN3SU
# HDSCD/AQ8rdHGO2n6Jl8P0zbr17C89XYcz1DTsEzOUyOArxCaC4Q6oRRRuLRvWoY
# WmEBc8pnol7XKHYC4jMYctenIPDC+hIK12NvDMk2ZItboKaDIV1fMHSRlJTYuVD5
# C4lh8zYGNRiER9vcG9H9stQcxWv2XFJRXRLbJbqvUAV6bMURHXLvjflSxIUXk8A8
# FdsaN8cIFRg/eKtFtvUeh17aj54WcmnGrnu3tz5q4i6tAgMBAAGjggHdMIIB2TAS
# BgkrBgEEAYI3FQEEBQIDAQABMCMGCSsGAQQBgjcVAgQWBBQqp1L+ZMSavoKRPEY1
# Kc8Q/y8E7jAdBgNVHQ4EFgQUn6cVXQBeYl2D9OXSZacbUzUZ6XIwXAYDVR0gBFUw
# UzBRBgwrBgEEAYI3TIN9AQEwQTA/BggrBgEFBQcCARYzaHR0cDovL3d3dy5taWNy
# b3NvZnQuY29tL3BraW9wcy9Eb2NzL1JlcG9zaXRvcnkuaHRtMBMGA1UdJQQMMAoG
# CCsGAQUFBwMIMBkGCSsGAQQBgjcUAgQMHgoAUwB1AGIAQwBBMAsGA1UdDwQEAwIB
# hjAPBgNVHRMBAf8EBTADAQH/MB8GA1UdIwQYMBaAFNX2VsuP6KJcYmjRPZSQW9fO
# mhjEMFYGA1UdHwRPME0wS6BJoEeGRWh0dHA6Ly9jcmwubWljcm9zb2Z0LmNvbS9w
# a2kvY3JsL3Byb2R1Y3RzL01pY1Jvb0NlckF1dF8yMDEwLTA2LTIzLmNybDBaBggr
# BgEFBQcBAQROMEwwSgYIKwYBBQUHMAKGPmh0dHA6Ly93d3cubWljcm9zb2Z0LmNv
# bS9wa2kvY2VydHMvTWljUm9vQ2VyQXV0XzIwMTAtMDYtMjMuY3J0MA0GCSqGSIb3
# DQEBCwUAA4ICAQCdVX38Kq3hLB9nATEkW+Geckv8qW/qXBS2Pk5HZHixBpOXPTEz
# tTnXwnE2P9pkbHzQdTltuw8x5MKP+2zRoZQYIu7pZmc6U03dmLq2HnjYNi6cqYJW
# AAOwBb6J6Gngugnue99qb74py27YP0h1AdkY3m2CDPVtI1TkeFN1JFe53Z/zjj3G
# 82jfZfakVqr3lbYoVSfQJL1AoL8ZthISEV09J+BAljis9/kpicO8F7BUhUKz/Aye
# ixmJ5/ALaoHCgRlCGVJ1ijbCHcNhcy4sa3tuPywJeBTpkbKpW99Jo3QMvOyRgNI9
# 5ko+ZjtPu4b6MhrZlvSP9pEB9s7GdP32THJvEKt1MMU0sHrYUP4KWN1APMdUbZ1j
# dEgssU5HLcEUBHG/ZPkkvnNtyo4JvbMBV0lUZNlz138eW0QBjloZkWsNn6Qo3GcZ
# KCS6OEuabvshVGtqRRFHqfG3rsjoiV5PndLQTHa1V1QJsWkBRH58oWFsc/4Ku+xB
# Zj1p/cvBQUl+fpO+y/g75LcVv7TOPqUxUYS8vwLBgqJ7Fx0ViY1w/ue10CgaiQuP
# Ntq6TPmb/wrpNPgkNWcr4A245oyZ1uEi6vAnQj0llOZ0dFtq0Z4+7X6gMTN9vMvp
# e784cETRkPHIqzqKOghif9lwY1NNje6CbaUFEMFxBmoQtB1VM1izoXBm8qGCAtcw
# ggJAAgEBMIIBAKGB2KSB1TCB0jELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hp
# bmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jw
# b3JhdGlvbjEtMCsGA1UECxMkTWljcm9zb2Z0IElyZWxhbmQgT3BlcmF0aW9ucyBM
# aW1pdGVkMSYwJAYDVQQLEx1UaGFsZXMgVFNTIEVTTjpGQzQxLTRCRDQtRDIyMDEl
# MCMGA1UEAxMcTWljcm9zb2Z0IFRpbWUtU3RhbXAgU2VydmljZaIjCgEBMAcGBSsO
# AwIaAxUAx2IeGHhk58MQkzzSWknGcLjfgTqggYMwgYCkfjB8MQswCQYDVQQGEwJV
# UzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UE
# ChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSYwJAYDVQQDEx1NaWNyb3NvZnQgVGlt
# ZS1TdGFtcCBQQ0EgMjAxMDANBgkqhkiG9w0BAQUFAAIFAOhXd2EwIhgPMjAyMzA3
# MTExNDM4MjVaGA8yMDIzMDcxMjE0MzgyNVowdzA9BgorBgEEAYRZCgQBMS8wLTAK
# AgUA6Fd3YQIBADAKAgEAAgIanQIB/zAHAgEAAgIR0TAKAgUA6FjI4QIBADA2Bgor
# BgEEAYRZCgQCMSgwJjAMBgorBgEEAYRZCgMCoAowCAIBAAIDB6EgoQowCAIBAAID
# AYagMA0GCSqGSIb3DQEBBQUAA4GBAL0G5QIvB9UVzd4j2ihfPXAVbzrSsy2fXkmG
# 47zzSqZNAME0CZdWrNCcRNR8L7azf0/cfLkRMpajZBwwQD21JkYijFBTpbByVezf
# 9JsRFZ4bEi24cqzEL2cen0443PtYXAbhlXjTNqTUu+0lns0EKLI+kXd9EjfulAMF
# BVQgQKfSMYIEDTCCBAkCAQEwgZMwfDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldh
# c2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBD
# b3Jwb3JhdGlvbjEmMCQGA1UEAxMdTWljcm9zb2Z0IFRpbWUtU3RhbXAgUENBIDIw
# MTACEzMAAAG59gANZVRPvAMAAQAAAbkwDQYJYIZIAWUDBAIBBQCgggFKMBoGCSqG
# SIb3DQEJAzENBgsqhkiG9w0BCRABBDAvBgkqhkiG9w0BCQQxIgQgkKiZnIAZZZVA
# DbNMwgKxo2B97h1f7eMDgpEnYzWwfx0wgfoGCyqGSIb3DQEJEAIvMYHqMIHnMIHk
# MIG9BCBk60bO8W85uTAfJVEO3vX2aLaQFcgcGpdwsOoi+foP9DCBmDCBgKR+MHwx
# CzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRt
# b25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xJjAkBgNVBAMTHU1p
# Y3Jvc29mdCBUaW1lLVN0YW1wIFBDQSAyMDEwAhMzAAABufYADWVUT7wDAAEAAAG5
# MCIEICR0gf5TSyyGAMIcnjJg8fTNR1yiM2K3JPNipHSlNX3EMA0GCSqGSIb3DQEB
# CwUABIICAJuNXuLbrXBT2h2ulsMlXTbrx6e6J0D9BFLtOWg/v/5PwMfiD7I20QYb
# rmWv7zd5A3Nb5yVrG0qOgTWXPDZo9IzmaTkTxAbzFcdb7sSldDlTbS9iqs0lEX4l
# voPXqjvGKhG4lB1laX/0Np21yXI9uvpZl7dll0oMpGVLNU5wbi9z1D8jmKCG4S6m
# yRpo182xhkoXeBIO3/g2jqhb4dkSw8Cp7Wo9tkOepdtvzTE90teXLqHVlIFm/zqc
# WwcbHwhnkEjAsvek2g30puNJHVD+w6G8L1nEaqpPGzobMZZFXCQj2HKxcfVMZz1H
# FvEeNf/rIhlNNzIOr27tCL9MOxzRyFRCg/2ZPQmDc9LcTKEUbwlbNeUBfgYB3QbO
# 0ClWh3HmAP3/9YRKQkrw+uTMeoo4TFU3YpDGoB6OuDnAdQpr6PIb7imePi7mIS8O
# L382BrcaAMDJzAl+nbZkeeboOqWX5Q5NzgMrmqA9KP53lpqF+LACqLXZjQgcV+5g
# HZRKyGjgHPuRig+PHSSOoctuk0QXdGBYkKHrHOtnRDq93nm5+T6LmF2XSH2LrWVz
# tvSdXCK5+skO8N7wPGa/jrVLWdW5uIdpg78egP0sfMgsG4rlqGmyJLwyLD7gX4b6
# Kor7NhvKakosGAn13CcE4p9ZMmMTVVySYz8mpQFHricwKOH2k/0a
# SIG # End signature block
