## ASR module
$script:versionId = "1.0011"

#registry key for ASR settings:
$script:ExploitGuardRegPath = "Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Policies\Microsoft\Windows Defender\Windows Defender Exploit Guard"
$script:ExploitGuardRegPrefPath = "Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows Defender\Windows Defender Exploit Guard"

# ASR rules name prefix
$script:settingNamePrefix = "Windows:Device:Policy.Config.Defender."

$timeLimitForSettingAction = 50 #in milliseconds. actions that take longer than that are logged.

# settingsNamesMapping - hash table
# Maps names of settings from policy to names used in cmdlet. 
# Also used as indicing for setting to unify usage throughout module.
$settingsNamesMapping = @{
    AttackSurfaceReductionRules = "AttackSurfaceReductionRules"
    AttackSurfaceReductionOnlyExclusions = "AttackSurfaceReductionOnlyExclusions"
    ControlledFolderAccessProtectedFolders = "ControlledFolderAccessProtectedFolders"
    ControlledFolderAccessAllowedApplications = "ControlledFolderAccessAllowedApplications"
    EnableControlledFolderAccess = "EnableControlledFolderAccess"
}

$unMergeableSettings = @(
    $settingsNamesMapping.EnableControlledFolderAccess
)

# regKeySuffixPerSetting - hash table
# Map of setting's index (cmdlet name from settingsNamesMapping) to registry key path suffix of general ExploitGuardRegPath. 
$script:regKeySuffixPerSetting = @{
    $settingsNamesMapping.AttackSurfaceReductionRules = "ASR\Rules"
    $settingsNamesMapping.AttackSurfaceReductionOnlyExclusions = "ASR\ASROnlyExclusions"
    $settingsNamesMapping.ControlledFolderAccessProtectedFolders = "Controlled Folder Access\ProtectedFolders"
    $settingsNamesMapping.ControlledFolderAccessAllowedApplications = "Controlled Folder Access\AllowedApplications"
    $settingsNamesMapping.EnableControlledFolderAccess = "Controlled Folder Access"
}
# groupsCounters - hash table
# maps setting name to counter of policies using it 
# used to get and report erros according to policy
$script:groupsCounters = @{
    $settingsNamesMapping.AttackSurfaceReductionRules = 0
    $settingsNamesMapping.AttackSurfaceReductionOnlyExclusions = 0
    $settingsNamesMapping.ControlledFolderAccessProtectedFolders = 0 
    $settingsNamesMapping.ControlledFolderAccessAllowedApplications = 0
}

# savedAsSeparateValuesInRegistry - array
# stored the settings that are saved in the same way in the registry:
# Their value is received as string of values delimitered by '|' and each single value is saved under the key as a different property name, with a value of DWORD '0'. 
$script:savedAsSeparateValuesInRegistry = @(
    $settingsNamesMapping.AttackSurfaceReductionOnlyExclusions,
    $settingsNamesMapping.ControlledFolderAccessProtectedFolders,
    $settingsNamesMapping.ControlledFolderAccessAllowedApplications
)

# index enum (for usage of $script:requestedValues)
if (-not ([System.Management.Automation.PSTypeName]'infoIndex').Type)
{
    Add-Type -TypeDefinition @"
        public enum infoIndex {
            action,
            groups,
            settingFirstErrorInfo
        }
"@ 
}

if (-not ([System.Management.Automation.PSTypeName]'firstErrorIndex').Type)
{
    Add-Type -TypeDefinition @"
        public enum firstErrorIndex {
            HR,
            errStr
        }
"@
}

Add-Type -TypeDefinition @"
    public enum registryType {
        GP,
        Preference
    }
"@

# requestedValues - hash table
# Map of setting's index (cmdlet name from settingsNamesMapping) to the values that need to be applied for each of them. 
# values are stored on Set call and applied only on CompletionActions (as those settings are mergeable and we need to store all the values before applying them).
$script:requestedValues = @{
    $settingsNamesMapping.AttackSurfaceReductionRules = @{} # map for rules to apply: @{ruleId = @(action,@(groups),@(HResult,ErroStr))}
    $settingsNamesMapping.AttackSurfaceReductionOnlyExclusions = @{} # map for exclusions to apply: @{exclustionValue =@(action,@(groups),@(HResult,ErroStr))}
    $settingsNamesMapping.ControlledFolderAccessProtectedFolders = @{} # map
    $settingsNamesMapping.ControlledFolderAccessAllowedApplications = @{} # map
}

# firstErrorInfo - hash table of setting to error information, in case there is no specific error in $script:requestedValues
$script:firstErrorInfo = @{}

# index enum (for usage of $rulesSupportInfo)
Add-Type -TypeDefinition @"
    public enum ruleSupportInfoIndex {
        S2012,
        S2016,
        S2019,
        Win10,
        Invalid
    }
"@

# rulesSupportInfo - hash table to validate if an ASR rule is supported on current machine
# ruleID = (2012 support, 2016 support, 2019 support ,min win10 version))
$rulesSupportInfo = @{
    "BE9BA2D9-53EA-4CDC-84E5-9B1EEEE46550" = @($true, $true, $true, 16299); # Block executable content from email client and webmail
    "D4F940AB-401B-4EFC-AADC-AD5F3C50688A" = @($true, $true, $true, 16299); # Block all Office applications from creating child processes
    "3B576869-A4EC-4529-8536-B80A7769E899" = @($true, $true, $true, 16299); #Block Office applications from creating executable content
    "75668C1F-73B5-4CF0-BB93-3ECF5CB7CC84" = @($true, $true, $true, 16299); #Block Office applications from injecting code into other processes
    "D3E037E1-3EB8-44C8-A917-57927947596D" = @($false, $false, $true, 16299); #Block JavaScript or VBScript from launching downloaded executable content
    "5BEB7EFE-FD9A-4556-801D-275E5FFC04CC" = @($true, $true, $true, 16299); #Block execution of potentially obfuscated scripts
    "92E97FA1-2EDF-4476-BDD6-9DD0B4DDDC7B" = @($false, $false, $true, 16299); #Block Win32 API calls from Office macros
    "01443614-cd74-433a-b99e-2ecdc07bfc25" = @($true, $true, $true, 17134); #Block executable files from running unless they meet a prevalence, age, or trusted list criterion
    "c1db55ab-c21a-4637-bb3f-a12568109d35" = @($true, $true, $true, 17134); #Use advanced protection against ransomware
    "9e6c4e1f-7d60-472f-ba1a-a39ef669e4b2" = @($true, $true, $true, 17134); #Block credential stealing from the Windows local security authority subsystem (lsass.exe)
    "d1e49aac-8f56-4280-b9ba-993a6d77406c" = @($true, $true, $true, 17134); #Block process creations originating from PSExec and WMI commands
    "b2b3f03d-6a65-4f7b-a9c7-1c7ef74a9ba4" = @($true, $true, $true, 16299); #Block untrusted and unsigned processes that run from USB
    "26190899-1602-49e8-8b27-eb1d0a1ce869" = @($true, $true, $true, 16299); #Block Office communication application from creating child processes
    "7674ba52-37eb-4a4f-a9a1-f0f9a1619a2c" = @($true, $true, $true, 17763); #Block Adobe Reader from creating child processes
    "e6db77e5-3df2-4cf1-b95a-636979351e5b" = @($false, $false, $true, 18362); #Block persistence through WMI event subscription
    "56a863a9-875e-4185-98a7-b882c64b5ce5" = @($true, $true, $true, 16299); #Block abuse of exploited vulnerable signed drivers
}

#array with info for rules validation (in case action is not supported).
$script:validRuleActions = @(0,1,2,6) # Disable (Disable the ASR rule),Block (Enable the ASR rule),Audit (Evaluate how the ASR rule would impact your organization if enabled),Warn (Enable the ASR rule but allow the end-user to bypass the block). Warn mode is available for most of the ASR rules.

#used to detect conflict and report accordingly for ASR rules
$script:conflictingRuleGroups = @()

#Dummy value for TriggerGPRefresh - adding and removing it from MpPreference
$script:dummyValue = "bfbfb459-07db-4fd1-a739-42c533f1119c-SIAM-7adcea70-53c6-43c6-86f0-83d6d5c00459"

####################################################### API functions #################################################################

function IsSupportedParam([string] $name)
{
    if (!($name.StartsWith($script:settingNamePrefix)))
    {
        return $false
    }

    $nameSuffix = GetNameSuffix $name

    return $settingsNamesMapping.ContainsKey($nameSuffix)
}

function Setup() 
{
    LogInfo (-join("ASR setup (", $script:versionId, ")"))

    #Import ConflictResolutionUtils
    try{
        $script:conflictUtilsModule = Import-Module -Name (Join-Path $global:WorkingFolder "ConflictResolutionUtils.psm1") -Force -AsCustomObject -PassThru -ErrorAction Stop
    } catch {
        LogWarning (-join("ASR setup: failed to import ConflictResolutionUtils module. HResult:",$_.Exception.HResult," error message:", $_.Exception.Message))
    }
    return
}

function IsApplicable([string] $name, [string] $value)
{
    try {
        $nameSuffix = GetNameSuffix $name

        # since rules are recieved as a string in one setting, we must check that they are all valid for current platform
        if($local:nameSuffix -eq $settingsNamesMapping.AttackSurfaceReductionRules)
        {
            #get this policy's requested rules for validation:           
            $singlePolicyRequestedRulesInfo = $script:conflictUtilsModule."ParseSettingInfo"($value)  # @{} ruleId, action        
            $invalidRulesIds = GetInvalidRules $local:singlePolicyRequestedRulesInfo
            if($local:invalidRulesIds.length -gt 0)
            {
                LogWarning (-join("ASR: at least one requested rule is not supported by this platform (OSVersion:",$global:OSVersion,", OSBuild:",$global:OSBuild,"): ", $invalidRulesIds))
                return $false
            }
        }

        if($nameSuffix -eq $settingsNamesMapping.EnableControlledFolderAccess)
        {
            if(($global:OSVersion -match $global:2012BuildStr) -or ($global:OSVersion -match $global:2016BuildStr)) #DL server
            {
                return $false
            }

            if([System.Convert]::ToInt32($global:OSBuild) -lt 16299) #old build of Windows 10 
            {
                return $false
            }
        }
    }
    catch {
        LogWarning (-join("ASR::IsApplicable failure for ",$local:name,", value:",$local:value,". HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
        return $false
    }
    return $true
}

function CompletionActions() 
{
    LogInfo "ASR: closing loose ends." 
    $startTimeMilli = GetCurrentTimeInMillis

    foreach($settingName in $settingsNamesMapping.Keys)
    {
        try {
            #FEIT1 - marker for tests: testing FirstErrorInfo - don't remove!

            # If set was called for this setting, clear prev content and apply the new value now.
            # If remove was called for this setting, then the value was already cleared in ResetToDefault
	   
            if(($script:requestedValues[$settingsNamesMapping.$local:settingName]).Count -gt 0)
            {
               LogInfo (-join("ASR: CompletionActions for ", $local:settingName))
               
               ClearCurrentASRContent -settingSuffix $local:settingName -regKeyPath (GetRegPathToKey ([registryType]::GP) $local:settingName) -value "" 
               
               # for mergeable settings (all settings, in this module), we must first remove and then apply, to match the (apply&remove conflict --> apply wins) schema in PolicyEnforcer
               # this shceme is maintained here because we first ResetToDef for all settings (Set only stores for further actions) and then we do CompletionActions which actually set.
               ApplySetting -settingSuffix $local:settingName
               # as a workaround for issue #39416943 of Defender, need to set and remove a dummy value for ASR Rules by Cmdlet to trigger the GP refresh
               TriggerGPRefresh $local:settingName
            }

        }
        catch
        {
            #if an unexpected exception happened, save it in the general error info
            if(!$script:firstErrorInfo.ContainsKey($settingsNamesMapping.$settingName))
                {
                    #no errors yet, add current error
                    $script:firstErrorInfo.Add($settingsNamesMapping.$settingName,( $_.Exception.HResult,$_.Exception.Message))
                    LogInfo (-join("ASR::CompletionActions 1st err rec for ", $settingName ,($script:firstErrorInfo[$settingsNamesMapping.$settingName] -join "|")))
                }
            LogWarning (-join("ASR::CompletionActions failure for ", $settingName ,". HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
            continue
        }
    }

    $totalTimeMilli = (GetCurrentTimeInMillis) - $startTimeMilli
    if($totalTimeMilli -gt $timeLimitForSettingAction)
    {
        LogInfo (-join("ASR: CompletionActions took ", $totalTimeMilli," ms."))
    }
}

function IsMergeSupported ([string] $name)
{
    $nameSuffix = GetNameSuffix $name

    return !($unMergeableSettings -contains $settingsNamesMapping.$nameSuffix)
}

#set assumes all rules are valid (otherwise IsApplicable would have failed the operation)
function Set([string] $name, [string] $value, [ref] $errorCode, [ref] $errorString) 
{
    $errorCode.Value = [HResult]::S_OK.value__
    $errorString.Value = ""
    $startTimeMilli = GetCurrentTimeInMillis

    $nameSuffix = GetNameSuffix $name

    LogInfo (-join("ASR: set value ", $nameSuffix, ", ", $value))

    try
    {
        switch ($settingsNamesMapping.$nameSuffix)
        {
            $settingsNamesMapping.EnableControlledFolderAccess {
                SetSingleValueToRegistry -regPath (GetRegPathToKey ([registryType]::GP) $nameSuffix) -name "EnableControlledFolderAccess" -type ([Microsoft.Win32.RegistryValueKind]::DWord) -value $value
                SetASRSettingByPref $_ @($value)
                break
            }
            $settingsNamesMapping.AttackSurfaceReductionRules {
                $script:groupsCounters[$settingsNamesMapping.AttackSurfaceReductionRules]++ #setting a new group of ASR rules
                #add rule to the "to be applied" map, because we might have several policies to apply.
                $singlePolicyRequestedRulesInfo = $script:conflictUtilsModule."ParseSettingInfo"($value)  # @{} ruleId, action
                $script:conflictUtilsModule."StoreSettingToRequested"($singlePolicyRequestedRulesInfo, $script:groupsCounters[$settingsNamesMapping.AttackSurfaceReductionRules], $script:requestedValues[$settingsNamesMapping.AttackSurfaceReductionRules], [ref]$script:conflictingRuleGroups)
                break
            }
            {$_ -in $script:savedAsSeparateValuesInRegistry } {
                $script:groupsCounters[$_]++ #setting a new group for current setting
                StoreNewValuesForApplication -settingSuffix $nameSuffix -value $value
                break
            }
            Default {
                #we are not supposed to get here since IsSupportedParam will fail an invalid setting name
                $logMessage = -join("ASR: invalid setting on Set: ", $nameSuffix)
                LogWarning $logMessage
                $errorCode.Value = [HResult]::E_NOINTERFACE.value__
                $errorString.Value = $logMessage
            }
        }
    }
    catch 
    {
        $errorCode.Value = $_.Exception.HResult
        $errorString.Value = $_.Exception.Message
        LogWarning (-join("ASR::Set failure for ",$nameSuffix,", value:",$value,". HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
    }

    $totalTimeMilli = (GetCurrentTimeInMillis) - $startTimeMilli
    if($totalTimeMilli -gt $timeLimitForSettingAction)
    {
        LogInfo (-join("ASR: Set for ",$nameSuffix," took ", $totalTimeMilli," ms."))
    }
}

function GetRegPathToKey([registryType] $regType, [string] $settingNameSuffix)
{
    switch ($regType)
    {
       ([registryType]::GP) { return (Join-Path $script:ExploitGuardRegPath $script:regKeySuffixPerSetting.($settingsNamesMapping.$settingNameSuffix)) }
       ([registryType]::Preference) { return (Join-Path $script:ExploitGuardRegPrefPath $script:regKeySuffixPerSetting.($settingsNamesMapping.$settingNameSuffix)) }
       Default { LogWarning "ASR GetRegPathToKey: invalid registry type received: $regType"; return $null}
    }
}

function ResetToDefault([string] $name, [string] $value, [ref] $errorCode, [ref] $errorString) 
{
    $errorCode.Value = [HResult]::S_OK.value__
    $errorString.Value = ""
    $startTimeMilli = ([TimeSpan](Get-Date -Format $global:TimeFormat)).TotalMilliseconds

    $nameSuffix = GetNameSuffix $name

    LogInfo (-join("ASR: ResToDef ",$nameSuffix," value ", $value))

    try
    {
        switch ($settingsNamesMapping.$nameSuffix)
        {
            $settingsNamesMapping.EnableControlledFolderAccess { 
               Remove-ItemProperty -Path (GetRegPathToKey ([registryType]::GP) $nameSuffix) -Name "EnableControlledFolderAccess" -ErrorAction SilentlyContinue #if item doesn't exits, continue.
               Set-MpPreference -EnableControlledFolderAccess Disabled #Workaround for bug #42275355. 
               break
            } 
            {$settingsNamesMapping.ContainsValue($_)}  { #for all other settings in this module, special handling:
                ClearCurrentASRContent -settingSuffix $nameSuffix -regKeyPath (GetRegPathToKey ([registryType]::GP) $nameSuffix) -value $value
                TriggerGPRefresh $_ # as a workaround for issue #39416943 of Defender, need to set and remove a dummy value for ASR settings by Cmdlet to trigger the GP refresh
                break
            }
            Default 
            {
                $errorCode.Value = [HResult]::E_NOINTERFACE.value__
                $errorString.Value = -join("ASR: ResetToDefault unsupported setting: ", $nameSuffix)
                LogWarning $errorString.Value
            }
        }
    }
    catch 
    {
        $errorCode.Value = $_.Exception.HResult
        $errorString.Value = $_.Exception.Message
        LogWarning (-join("ASR::ResToDef failure for ",$nameSuffix,", value:",$value,". HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
    }

    $totalTimeMilli = (GetCurrentTimeInMillis) - $startTimeMilli
    if($totalTimeMilli -gt $timeLimitForSettingAction)
    {
        LogInfo (-join("ASR: ResToDef for ",$nameSuffix," took ", $totalTimeMilli," ms."))
    }
}

function VerifyAssignment([string] $name, [string] $value, [ref] $enforceSucceeded, [ref] $errorCode, [ref] $errorString)
{
    $errorCode.Value = [HResult]::S_OK.value__
    $errorString.Value = ""
    $startTimeMilli = (GetCurrentTimeInMillis)

    $nameSuffix = GetNameSuffix $name

    LogInfo (-join("ASR: VerAssign ",$nameSuffix," value ", $value))

    $currentValueIfNotEqual = $null
    $enforceSucceeded.Value = $true
    $conflictDetected = $false
    $settingErrorInformation = @()
    $result = $true
    try
    {
        $mpPref = Get-MpPreference
        if($null -eq $mpPref)
        {
            $enforceSucceeded.Value = $false
            $errorString.Value = "Failed to Get-MpPreference for validation."
            LogWarning $errorString.Value
            return [HResult]::E_FAIL
        }

        switch ($settingsNamesMapping.$nameSuffix)
        {
            $settingsNamesMapping.EnableControlledFolderAccess {
                $currentValue = $mpPref.EnableControlledFolderAccess
                if($currentValue -ne $value)
                {
                    $result = $false
                    $currentValueIfNotEqual = $currentValue
                }
                break
            }
            $settingsNamesMapping.AttackSurfaceReductionRules {

                $singlePolicyRequestedRulesInfo = $script:conflictUtilsModule."ParseSettingInfo"($value)
                $rulesArr = @($local:singlePolicyRequestedRulesInfo.Keys)
                $groups = $script:conflictUtilsModule."GetGroupFromSettings"($local:rulesArr, $script:requestedValues[$settingsNamesMapping.AttackSurfaceReductionRules])
                # groups contain all the policies that have the SAME set of rules. So we can check
                # for conflict based on the 1st one because they must all have the same conflict result if they have the same rules in them.
                $groupForValidation = $local:groups[0] 
                
                if($script:conflictingRuleGroups -contains $groupForValidation) #check if this group is in a conflict
                {
                    LogInfo (-join("ASR:VerAssign: detected conflict based on group:",$groupForValidation))
                    $conflictDetected = $true
                    $result = $false
                }
                else
                {
                    LogInfo (-join("ASR:VerAssign: validation for group:",$groupForValidation))
                    $currentRulesIDsArr = @($mpPref.AttackSurfaceReductionRules_Ids)
                    $currentRulesActionsArr = @($mpPref.AttackSurfaceReductionRules_Actions) #returns 1 and not 'Enabled'
                    $result = validateASRRulesAssignment -currentRulesIds $currentRulesIDsArr -currentActions $currentRulesActionsArr -requestedRulesInfo $singlePolicyRequestedRulesInfo
                    if(!$result) #adding debug traces for bug #43959951
                    {
                        LogDebug (-join("ASR:VerAssign: failed to verify ASR rules. Got from MpPref IDs array:",($currentRulesIDsArr -join("|")) ,", Actions array:",($currentRulesActionsArr -join("|"))))
                    }
                }
                
                if(!$result) #detected assignment was not done
                {
                    $currentValueIfNotEqual = ConstructCurrentValue -currentRulesIds $currentRulesIDsArr -currentActions $currentRulesActionsArr
                    $settingErrorInformation = $script:conflictUtilsModule."GetFirstErrorInfoByGroup"($groupForValidation, $script:requestedValues[$settingsNamesMapping.AttackSurfaceReductionRules])
                    if($currentValueIfNotEqual -eq "=") #adding debug traces for bug #43959951
                    {
                        LogDebug (-join("ASR:VerAssign: current value for ASR rules is empty, trying to fetch it again"))
                        $mpPref = Get-MpPreference
                        $currentRulesIDsArr = @($mpPref.AttackSurfaceReductionRules_Ids)
                        $currentRulesActionsArr = @($mpPref.AttackSurfaceReductionRules_Actions)
                        LogDebug (-join("ASR:VerAssign: On second try, got from MpPref IDs array:",($currentRulesIDsArr -join("|")) ,", Actions array:",($currentRulesActionsArr -join("|"))))
                        $regValue = Get-ItemProperty -Path (GetRegPathToKey ([registryType]::GP) 'AttackSurfaceReductionRules') -Name $singlePolicyRequestedRulesInfo[0] #getting one rule from registry
                        LogDebug (-join("ASR:VerAssign: Got from registry value:",$regValue))
                    }
                }
                break
            }
            {$_ -in $script:savedAsSeparateValuesInRegistry } {
                $currentValuesArr = GetMpProperty -setting $nameSuffix -mpPrefObj $mpPref 
                $result = validateValuesAssignment -currentValues $currentValuesArr -requestedValues ($value.Split("|") | where {$_})
                
                if(!$result) #These settings are mergeable and therefore should use Internal Conflict Resolution module even though they are non-conflict, they use the same structs (groups, errorInfo) to detect what policy they come from
                {
                    $currentValueIfNotEqual = $currentValuesArr -join "|"
                    $groupId = ($script:conflictUtilsModule."GetGroupIdFromValue"($value, $script:groupsCounters[$settingsNamesMapping.$nameSuffix], $script:requestedValues[$settingsNamesMapping.$nameSuffix]))[0]
                    $settingErrorInformation =  $script:conflictUtilsModule."GetFirstErrorInfoByGroup"($groupId, $script:requestedValues[$settingsNamesMapping.$nameSuffix])
                }
                break
            }
            Default {
                #we are not supposed to get here since IsSupportedParam will fail an invalid setting name
                $logMessage = -join("ASR: invalid setting on VerAssign: ", $nameSuffix)
                LogInfo $logMessage
                $errorCode.Value = [HResult]::E_NOINTERFACE.value__
                $errorString.Value = $logMessage
                $result = $false
            }
        }
    }
    catch 
    {
        $errorCode.Value = $_.Exception.HResult
        $errorString.Value = $_.Exception.Message
        LogWarning (-join("ASR: VerAssign failure for ", $nameSuffix, ", value:",$value,". HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
        return [HResult]::E_Fail 
    }

    if(!$result)  
    {
        #assignemt failed, output info
        $enforceSucceeded.Value = $false
        LogWarning (-join("ASR:VerAssign returns false for ",$nameSuffix," (current:", $currentValueIfNotEqual, ", expected:",$value,")"))

        #get errors information, if exist
        if($local:conflictDetected)
        {
            $errorCode.Value = [HResult]::E_INTERNAL_CONFLICT
            $errorString.Value = "conflict detected for requested value" 
        } 
        elseif($settingErrorInformation.Count -gt 0) 
        {
            $errorCode.Value = $settingErrorInformation[0]
            $errorString.Value = $settingErrorInformation[1]
        }
        elseif($script:firstErrorInfo.ContainsKey($settingsNamesMapping.$nameSuffix))
        {
            #TBD: get error for relevant group?
            $errorCode.Value = $script:firstErrorInfo[$settingsNamesMapping.$nameSuffix][0]
            $errorString.Value = $script:firstErrorInfo[$settingsNamesMapping.$nameSuffix][1]
        }
    }

    $totalTimeMilli = (([TimeSpan](Get-Date -Format $global:TimeFormat)).TotalMilliseconds) - $startTimeMilli
    if($totalTimeMilli -gt $timeLimitForSettingAction)
    {
        LogInfo (-join("ASR: VerAssign for ",$nameSuffix,", " , $value, " took ", $totalTimeMilli," ms."))
    }

    return [HResult]::S_OK
}

function VerifyRemoval([string] $name, [string] $value, [ref] $enforceSucceeded, [ref] $errorCode, [ref] $errorString)
{
    $errorCode.Value = [HResult]::S_OK.value__
    $errorString.Value = ""
    $startTimeMilli = ([TimeSpan](Get-Date -Format $global:TimeFormat)).TotalMilliseconds

    $nameSuffix = GetNameSuffix $name

    LogInfo (-join("ASR: VerRemov ",$nameSuffix," value ", $value))
    $enforceSucceeded.Value = $true
    try
    {
        switch ($settingsNamesMapping.$nameSuffix)
        {
            $settingsNamesMapping.EnableControlledFolderAccess {
                $currentGPRegItem = (GetSingleCurrentValueFromRegistry -registryPath (GetRegPathToKey ([registryType]::GP) $nameSuffix) -valueName "EnableControlledFolderAccess")
                $currentPrefRegItem = (GetSingleCurrentValueFromRegistry -registryPath (GetRegPathToKey ([registryType]::Preference) $nameSuffix) -valueName "EnableControlledFolderAccess")
                
                $GPHiveValueCleared = (($null -eq $currentGPRegItem) -or ($currentGPRegItem.Length -eq 0))
                $PrefHiveValueCleared = (($null -eq $currentPrefRegItem) -or ($currentPrefRegItem.Length -eq 0) -or ($currentPrefRegItem -eq 0)) #Last option is part of the workaround for bug #42275355.
                if (!$PrefHiveValueCleared -or !$GPHiveValueCleared) # a value exists in registry
                {        
                    LogWarning (-join("ASR: VerRemov ",$local:nameSuffix," value ",$local:value," not removed. values GP:[",$currentGPRegItem,"] Pref:[",$currentPrefRegItem,"]"))
                    $enforceSucceeded.Value = $false    
                }
                break
            }
            {$settingsNamesMapping.AttackSurfaceReductionRules -or ($_ -in $script:savedAsSeparateValuesInRegistry )} {

                $enforceSucceeded.Value = ValidateMergableSettingRemoval $local:nameSuffix $local:value 
                break
            }
            Default {
                #we are not supposed to get here since IsSupportedParam will fail an invalid setting name
                $logMessage = -join("ASR: invalid setting on VerRemov: ", $nameSuffix)
                LogInfo $logMessage
                $errorCode.Value = [HResult]::E_NOINTERFACE.value__
                $errorString.Value = $logMessage
            }
        }
    }
    catch 
    {
        $errorCode.Value = $_.Exception.HResult
        $errorString.Value = $_.Exception.Message
        LogWarning (-join("ASR: VerRemov failure for ", $nameSuffix, ", value:",$value,". HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
        return [HResult]::E_Fail 
    }

    if($enforceSucceeded.Value -eq $false)
    {
        #get errors information, if exist
        if($script:firstErrorInfo.ContainsKey($settingsNamesMapping.$nameSuffix))
        {
            $errorCode.Value = $script:firstErrorInfo[$settingsNamesMapping.$nameSuffix][[firstErrorIndex]::HR]
            $errorString.Value = $script:firstErrorInfo[$settingsNamesMapping.$nameSuffix][[firstErrorIndex]::errStr]
        }
    }

    $totalTimeMilli = (([TimeSpan](Get-Date -Format $global:TimeFormat)).TotalMilliseconds) - $startTimeMilli
    if($totalTimeMilli -gt $timeLimitForSettingAction)
    {
        LogInfo (-join("ASR: VerRemov for ",$nameSuffix,", " , $value, " took ", $totalTimeMilli," ms."))
    }

    return [HResult]::S_OK
}

 ####################################################### utility functions: #######################################################

 # ValidateMergableSettingRemoval - 
 # validate setting (of type mergeable, including ASR rules), were removed from 
 # both GP and Pref hives. 
 # NOTE: if $valueToRemove is $null (we cannot validate for specific values), 
 # function will validate that there are no unexpected values in GP registry
 # (not in prefs hive because there might be local admin settings there as well as our settings)
 # RETURN: $true in the following cases: 
 # (a) $valueToRemove were removed.
 # (b) $valueToRemove is $null, but there are no "unexpected" values in the GP hive.
 # otherwise - $false.
 function ValidateMergableSettingRemoval([string]$settingSuffix, [string]$valueToRemove)
 {
    $valuesRemovedArray = $local:valueToRemove.Split("|") | where {$_}
    
    #cannot use -AdditionalChildPath because it is not supported on older PS versions
    $GPkeyPath = (GetRegPathToKey ([registryType]::GP) $settingSuffix)
    $PrefkeyPath = (GetRegPathToKey ([registryType]::Preference) $settingSuffix)

    if(($null -eq $local:valuesRemovedArray) -or ($local:valuesRemovedArray.Count -eq 0)) 
    {
        return ValidateRegistryCurrentValueWasRequested -settingSuffix $local:settingSuffix -regPath $local:GPkeyPath
    }

    # $valueToRemove was not $null, need to validate specific content removal
    return ValidateContentRemovalFromRegistries -settingSuffix $local:settingSuffix -contentRemoved $local:valuesRemovedArray -GPKeyPath $local:GPKeyPath -PrefkeyPath $local:PrefkeyPath

 }

 # ValidateRegistryCurrentValueWasRequested - 
 # Function returns $true if all the values in the requested hive were requested for Set.
 function ValidateRegistryCurrentValueWasRequested([string]$settingSuffix, [string]$regPath)
 {
    $currentGPRegItem = GetCurrentValueFromReg $local:settingSuffix $local:regPath 

    foreach($singleValue in $local:currentGPRegItem)
    {
        if(!(($script:requestedValues[$settingsNamesMapping.$local:settingSuffix]).ContainsKey($local:singleValue)))
        {
            LogWarning (-join("ASR:ValidateRegistryCurrentValueWasRequested for ",$local:settingSuffix," detected value (",$local:singleValue,") in GP hive that was not requested. Assuming clear failed"))
            return $false
        }
    }

    # all rules in registry were requested (or registry is empty). Assuming clear was successful.
    LogInfo (-join("ASR:ValidateRegistryCurrentValueWasRequested for ",$local:settingSuffix,", values in registry (if any), were requested."))
    return $true
 }

 # ValidateContentRemovalFromRegistries - 
 # function returns $true if:
 # (a) $contentRemoved does NOT exist in GP and Prefs registry hives. 
 # (b) $contentRemoved (or some of it) exists in GP&Pref hives, but it was also requested for Set.
 function ValidateContentRemovalFromRegistries ([string]$settingSuffix, [string[]]$contentRemoved, [string]$GPKeyPath, [string]$PrefkeyPath)
 {
    foreach($singleValue in $local:contentRemoved)
    {
        #extract first part of the value (rule name, or exclusion name).
        $valueInfoArr = $singleValue.Split("=") 
        $valueName = $valueInfoArr[0]

        $GPItem = Get-Item -Path $local:GPKeyPath
        $PrefItem = Get-Item -Path $local:PrefkeyPath
        $entryKeyPropertyExistsInGP = ($local:valueName -in ($local:GPItem.Property))
        $entryKeyPropertyExistsInPref = ($local:valueName -in ($local:PrefItem.Property)) 
        $valueIsRequestedForApply = ($script:requestedValues[$settingsNamesMapping.$local:settingSuffix]).ContainsKey($local:valueName)

        LogInfo (-join("ASR:ValContentRemFromRegs for ",$local:settingSuffix," value:",$local:valueName,". In GP:[",(($local:GPItem.Property) -join("|")),"] (",$entryKeyPropertyExistsInGP,"). In Pref:[",(($local:PrefItem.Property) -join ("|")),"] (",$entryKeyPropertyExistsInPref,"). Is requested:",$valueIsRequestedForApply))

        if(($local:valueIsRequestedForApply) -or (!$local:entryKeyPropertyExistsInGP -and !$local:entryKeyPropertyExistsInPref))
        {
            #$singleValue was removed (or it is still needed), continue checking other values 
            continue
        }

        if($local:entryKeyPropertyExistsInGP) 
        {
            LogWarning (-join("ASR:VerRem for value:",$local:valueName," GP entry [",$local:GPKeyPath. "] exists and was not requested:",$local:valueIsRequestedForApply))
        }
        if($local:entryKeyPropertyExistsInPref)
        {
            LogWarning (-join("ASR:IsDefined for value:",$local:valueName," check entry [",$local:PrefkeyPath. "] exits, was not requested(",$local:valueIsRequestedForApply,")"))
        }

        #if we found a single rule that was not applied, we will report failure, no need to chech the others in the same policy.
        return $false
    }
                
    return $true
 }

 # GetSingleCurrentValueFromRegistry - returns the value of a specific 
 # registry key at $registryPath, with the name $valueName
 # If no such key exists, returns $null.
  function GetSingleCurrentValueFromRegistry([string] $registryPath, [string]$valueName)
 {
    $registryItemObj = Get-ItemProperty $registryPath $valueName -ErrorAction SilentlyContinue
    if($registryItemObj)
    {
        return $registryItemObj.($valueName)
    }
    return $null
 }

 # ApplySetting - do the actual application of the setting we have 
 # consumed at Set(). Uses the script:requestedValues map.
 # If catches error - error info is stored in script:requestedValues and will be 
 # available when validation of setting needs it.
 function ApplySetting([string] $settingSuffix)
{
    try {
        #FEIT2 - marker for tests: testing FirstErrorInfo - don't remove!

        $regKeyPath =  (GetRegPathToKey ([registryType]::GP) $local:settingSuffix)
        
        # this is out of the switch because it is common for all ASR settings when using GP hive
        if(!(Test-Path $regKeyPath)) 
        {
            LogInfo (-join("ASR: ApplySetting for:", $settingSuffix, ", reg key doesn't exist. creating: ", $regKeyPath))
            try {
                New-Item -Path $regKeyPath -Force -ErrorAction Stop | Out-Null
            } 
            catch {
                LogWarning (-join("ASR: ApplySetting failed to create reg key:", $regKeyPath, ". HResult:",$_.Exception.HResult,", error message:",$_.Exception.Message))
            }
        }

        switch($settingsNamesMapping.$settingSuffix)
        {
            $settingsNamesMapping.AttackSurfaceReductionRules 
            {
                # add ASR rules that should be active
                SetNewASRRules
                break
            }
            {$_ -in $script:savedAsSeparateValuesInRegistry }
            {
                # add ASR exclusions that should be active
                SetASRSetting -settingSuffix $settingSuffix -regPath $regKeyPath
                break
            }
            Default
            {
                LogWarning (-join("ASR::ApplySetting invalid settingSuffix:",$settingSuffix,". Nothing is applied."))
            }
        }
    }
    catch
    {
        if(!$script:firstErrorInfo.ContainsKey($settingsNamesMapping.$settingSuffix))
        {
            #no errors yet, add current error
            $script:firstErrorInfo.Add($settingsNamesMapping.$settingSuffix,( $_.Exception.HResult,$_.Exception.Message))
            LogInfo (-join("ASR::ApplySetting 1st err rec for ", $settingSuffix ,($script:firstErrorInfo[$settingsNamesMapping.$settingSuffix] -join "|")))

        }
        LogWarning (-join("ASR::ApplySetting failure for ", $settingSuffix ,". HResult:",$_.Exception.HResult,", error message:",$_.Exception.Message))
    }
}

 # StoreNewValuesForApplication - store value (NOT ASR rules), for later application,
 # in script:requestedValues  (value & group)
function StoreNewValuesForApplication([string] $settingSuffix, [string] $value)
{
    $newValuesArr = $value.Split("|") | where {$_}

    LogDebug (-join("ASR: StoreNewValuesForApplication newValuesArr:[", ($newValuesArr -join "|"),"]"))

    foreach($singleValue in $newValuesArr)
    {
        if(($script:requestedValues[$settingsNamesMapping.$settingSuffix]).ContainsKey($singleValue))
        {
            #value already requested by another policy, add new group to it
            ($script:requestedValues[$settingsNamesMapping.$settingSuffix])[$singleValue][([infoIndex]::groups)] += $script:groupsCounters[$settingsNamesMapping.$settingSuffix]
            LogDebug (-join("ASR: new value ",$singleValue ," matches one that exists, adding group. {",$singleValue,"=(0,",(($script:requestedValues[$settingsNamesMapping.$settingSuffix])[$singleValue][([infoIndex]::groups)] -join("|")),"),()]")) 
        }
        else
        {
            #value asked for the 1st time - add it
            LogDebug (-join("ASR: added new value to be set for ", $settingSuffix,": {",$singleValue,"=(0,",$script:groupsCounters[$settingsNamesMapping.$settingSuffix],"),()]")) 
            ($script:requestedValues[$settingsNamesMapping.$settingSuffix]).Add($singleValue,@(0,@($script:groupsCounters[$settingsNamesMapping.$settingSuffix]),@()))
        }
    }
}

 # StoreASRRulesToRequestedRules - store ASR rules values (id & action & group), for later application,
 # in script:requestedValues
 function StoreASRRulesToRequestedRules ([hashtable] $requestedRulesInfoToAdd)
 {
    $foundConflictForThisGroup = $false

    foreach($ruleId in $requestedRulesInfoToAdd.keys)
    {
        $currentGroup = $script:groupsCounters[$settingsNamesMapping.AttackSurfaceReductionRules]
        #rule is already requested
        if(($script:requestedValues[$settingsNamesMapping.AttackSurfaceReductionRules]).ContainsKey($ruleId))
        {
            #in any case - add new groupID to this rule
            ($script:requestedValues[$settingsNamesMapping.AttackSurfaceReductionRules])[$ruleId][([infoIndex]::groups)] += $currentGroup

            #if applied rule's action doesn't match on several policies, report conflict
            if(!$foundConflictForThisGroup -and ($requestedRulesInfoToAdd[$ruleId] -ne ($script:requestedValues[$settingsNamesMapping.AttackSurfaceReductionRules])[$ruleId][([infoIndex]::action)]))
            {
                $errorMsg = (-join("ASR: conflict recognized for rule (",$ruleId,"). actions: ",($script:requestedValues[$settingsNamesMapping.AttackSurfaceReductionRules])[$ruleId][([infoIndex]::action)],",",$requestedRulesInfoToAdd[$ruleId]))
                LogWarning $errorMsg

                #mark all ASR rules previously stored for application since we found a conflict and will not set them.
                $conflictingGroups = ($script:requestedValues[$settingsNamesMapping.AttackSurfaceReductionRules])[$ruleId][([infoIndex]::groups)] + $currentGroup 
                addConflictingGroups -conflictingGroups $conflictingGroups
                LogDebug (-join("ASR: currently conflicting groups:",($conflictingGroups -join("|")),". Total conflicting groups:",($script:conflictingRuleGroups -join("|"))))
            } 

        }
        else
        {
            #first time we see this rule - add new rule and action (no error, yet)
            LogDebug (-join("ASR: adding new rule to be set: [",$ruleId,",",$requestedRulesInfoToAdd[$ruleId],",",$currentGroup,"]"))
            ($script:requestedValues[$settingsNamesMapping.AttackSurfaceReductionRules]).Add($ruleId,@($requestedRulesInfoToAdd[$ruleId],@($currentGroup),@()))
        }
    }
 }

  # GetMpProperty - returns the setting specific ouput of Get-MpPreference. 
  # input: [Object]$mpPrefObj - is the output object of Get-MpPreference
  function GetMpProperty ([string]$setting, [Object]$mpPrefObj)
 {
    $mpPreferenceProperty = $null

    switch($settingsNamesMapping.$setting)
    {
        $settingsNamesMapping.AttackSurfaceReductionOnlyExclusions {
            $mpPreferenceProperty = $mpPrefObj.AttackSurfaceReductionOnlyExclusions
            break
        }
        $settingsNamesMapping.ControlledFolderAccessProtectedFolders {
            $mpPreferenceProperty = $mpPrefObj.ControlledFolderAccessProtectedFolders
            break
        }
        $settingsNamesMapping.ControlledFolderAccessAllowedApplications {
            $mpPreferenceProperty = $mpPrefObj.ControlledFolderAccessAllowedApplications
            break
        }
        Default {
            LogWarning (-join("ASR:GetMpProperty invalid setting: ",$setting))
        }
    }
    $mpPreferenceProperty = $mpPreferenceProperty.ForEach({ $_.Trim() }) #in DL servers Get-MpPreference output has spaces in the values, so we need to remove them for validateValuesAssignment
    return $mpPreferenceProperty
 }

# validateValuesAssignment - (NOT ASR rules) validates that values requested (uses script:requestedValues)
# exist in the values array provided.
 function validateValuesAssignment ([string[]] $currentValues, [string[]] $requestedValues)
 {
    foreach($singleVal in $requestedValues)
    {
        $singleVal = $singleVal.Trim() #Spaces are removed from values in Get-MpPreference output in non-DL servers
        if(!($singleVal -in $currentValues))
        {
            return $false
        }
    }

    return $true
 }

 # ConstructCurrentValue - use Get-MpPreference output of ASR rules to return a user-formatted
 # current value string.
 function ConstructCurrentValue ([string[]] $currentRulesIds, [string[]] $currentActions)
 {
    $currentValueStr = ""

    for(($i=0); ($i -lt $currentRulesIds.Count); ($i++))
    {   
        $singleRule = $currentRulesIds[$i] + "=" + $currentActions[$i]

        if($i -gt 0)
        {
            $currentValueStr += "|"
        }

        $currentValueStr += $singleRule
    }

    return $currentValueStr
 }

 # validateASRRulesAssignment - (only ASR rules) validates that rules requested ($requestedRulesInfo)
# exist in the values array provided ($currentRulesIds & $currentActions, from the Get-MpPreference output).
 function validateASRRulesAssignment ([string[]] $currentRulesIds, [string[]] $currentActions, [hashtable] $requestedRulesInfo)
 {
    foreach ($ruleId  in @($requestedRulesInfo.keys))
    {
        # get index of rule and action
        $index = (0..($currentRulesIds.Count-1)) | where {$currentRulesIds[$_] -eq $ruleId}

        if($null -eq $index)
        {
            # requested rule was not set
            return $false 
        }

        if($currentActions[$index] -ne $requestedRulesInfo[$ruleId])
        {
            # requested action for this rule was not set
            return $false 
        }
    }

    return $true
 }

# SetASRSetting - (NOT ASR rules) sets the values for setting in the registry location provided.
# uses script:requestedValues to get the values, after they were set in Set().
# NOTE: assumes registry path exists.
function SetASRSetting([string] $settingSuffix, [string] $regPath)
{
    LogDebug (-join("ASR: SetASRSetting: ", $settingSuffix, " values:",(($script:requestedValues[$settingsNamesMapping.$settingSuffix]).keys -join("|"))))

    foreach($singleName in ($script:requestedValues[$settingsNamesMapping.$settingSuffix]).keys)
    {      
        try
        {
            #FEIT4 - marker for tests: testing FirstErrorInfo - don't remove!
            
            # try-catch inside loop to allow max possible assignment -
            # if assignment fails, we save the error and continue with the assignment
            # in case of false on VerifyAssignment we will check if we have an error to show.
            LogDebug (-join("ASR: SetASRSetting setting ", $settingSuffix," value:",$singleName))
            
            # when GP sets exclusions the value is set to 0 and ignored. If fail on creation, we will discover and report on IsEqual.
            SetSingleValueToRegistry -regPath $regPath -name $singleName -value "0" -type ([Microsoft.Win32.RegistryValueKind]::String)
        }
        catch
        {
            $script:requestedValues[$settingsNamesMapping.$settingSuffix][$singleName][[infoIndex]::settingFirstErrorInfo] = @($_.Exception.HResult,$_.Exception.Message)
            LogWarning (-join("ASR::SetASRSetting failure for ", $singleName ," HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
            continue
        }
    }
}

# as a workaround for issue #39416943 of Defender, need to set and remove a dummy value for AV exclusions by Cmdlet to trigger the GP refresh
function TriggerGPRefresh([string] $setting)
{
    switch($local:setting)
    {
        $settingsNamesMapping.AttackSurfaceReductionOnlyExclusions { 
            Add-MpPreference -AttackSurfaceReductionOnlyExclusions $script:dummyValue
            Remove-MpPreference -AttackSurfaceReductionOnlyExclusions $script:dummyValue
            break
        }
        $settingsNamesMapping.ControlledFolderAccessProtectedFolders { 
            Add-MpPreference -ControlledFolderAccessProtectedFolders $script:dummyValue
            Remove-MpPreference -ControlledFolderAccessProtectedFolders $script:dummyValue
            break
        } 
        $settingsNamesMapping.ControlledFolderAccessAllowedApplications { 
            Add-MpPreference -ControlledFolderAccessAllowedApplications $script:dummyValue
            Remove-MpPreference -ControlledFolderAccessAllowedApplications $script:dummyValue
            break
        }
        $settingsNamesMapping.AttackSurfaceReductionRules { 
            Add-MpPreference -AttackSurfaceReductionRules_Ids $script:dummyValue -AttackSurfaceReductionRules_Actions 1
            Remove-MpPreference -AttackSurfaceReductionRules_Ids $script:dummyValue -AttackSurfaceReductionRules_Actions 1
            break
        }
        default { LogWarning (-join("ASR::TriggerGPRefresh: invalid setting name ",$local:setting))}
    }
}

function SetASRSettingByPref ([string]$nameSuffix, [string[]]$values)
{
    Switch($nameSuffix)
    {
        $settingsNamesMapping.EnableControlledFolderAccess { Set-MpPreference -EnableControlledFolderAccess $local:values[0]; break}
        Default {
            LogWarning (-join("ASR::SetASRSettingByPref no cmdlet for ",$nameSuffix))}
    }
}

 # SetSingleValueToRegistry - sets specified $value (of type $type) to the registry key specified
 # by $regPath, with the key name of $name. if no such registry path, function will create it.
 function SetSingleValueToRegistry([string] $regPath, [string] $name, [string] $type, [string] $value)
{
    if(!(Test-Path $regPath)) 
    {
        LogInfo (-join("ASR SetSingleValueToRegistry: reg key ",$regPath, " doesn't exist. creating."))
        New-Item -Path $regPath -Force | Out-Null
    }
    Set-ItemProperty -Path $regPath -Name $name -Type $type -Value $value -ErrorAction Stop #SilentlyContinue | Out-Null 
}

 # IsRuleFromGroupWithNoConflict - return true if this rule is a part of a policy which doesn't have a conflict
 # In other words: this is a rule that should be applied
 function IsRuleFromGroupWithNoConflict([string] $ruleId)
 {
    return(!($script:conflictUtilsModule."IsSubSettingOnlyFromGroupInConflict"($ruleId, $script:requestedValues[$settingsNamesMapping.AttackSurfaceReductionRules], $script:conflictingRuleGroups))) #$ruleInAConflictingGroup
 }

  # SetNewASRRules - set ASR rules (uses script:requestedValues)
 function SetNewASRRules()
 { 
    foreach($ruleId in ($script:requestedValues[$settingsNamesMapping.AttackSurfaceReductionRules]).keys)
    {       
        if(IsRuleFromGroupWithNoConflict $ruleId)
        {
            LogInfo (-join("ASR SetNewASRRules for ",$ruleId,"."))
            $action = ($script:requestedValues[$settingsNamesMapping.AttackSurfaceReductionRules])[$ruleId][([infoIndex]::action)]
            try
            {
                #FEIT3 - marker for tests: testing FirstErrorInfo - don't remove!
                LogInfo (-join("ASR::SetNewASRRules using SetSingleValueToRegistry for", $local:ruleId ," with action:", $local:action))
                SetSingleValueToRegistry -regPath (GetRegPathToKey ([registryType]::GP) 'AttackSurfaceReductionRules') -name $local:ruleId -value $local:action -type ([Microsoft.Win32.RegistryValueKind]::DWord)
            }
            catch
            {
                $script:requestedValues[$settingsNamesMapping.AttackSurfaceReductionRules][$ruleId][[infoIndex]::settingFirstErrorInfo] = @($_.Exception.HResult,$_.Exception.Message)
                LogWarning (-join("ASR::SetNewASRRules failure for ", $ruleId ," action:", $action ,". HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
            }
        }
        else
        {
            LogInfo (-join("ASR SetNewASRRules conflict for ",$ruleId," - not applying."))
        }
    }
 }

# isRuleIdSupported - return true if ASR rule is supported on this platform
  function isRuleIdSupported([string] $ruleId){
        
     if(!$rulesSupportInfo.ContainsKey($ruleId))
     {
        return $false
     }

     $index = GetPlatformIndexByMachinePlatformStr
     if($index -eq [ruleSupportInfoIndex]::Invalid)
     {
        return $false
     }

     if($index -eq [ruleSupportInfoIndex]::Win10)
     {
        return ($global:OSBuild -ge $rulesSupportInfo[$ruleId][$index])
     }

     return $rulesSupportInfo[$ruleId][$index]
 }

 function GetPlatformIndexByMachinePlatformStr()
 {
    switch($global:OSVersion) { 
        {$_ -match $global:2012BuildStr} {return [ruleSupportInfoIndex]::S2012}
        {$_ -match $global:2016BuildStr} {return [ruleSupportInfoIndex]::S2016}
        {($_ -match $global:2019BuildStr) -or ($_ -match $global:2022BuildStr)} {return [ruleSupportInfoIndex]::S2019}
        {($_ -match $global:Win10Str) -or ($_ -match $global:Win11Str)} {return [ruleSupportInfoIndex]::Win10} #also for Win11
        Default {return [ruleSupportInfoIndex]::Invalid} 
    }
 }

 function isActionSupported([int] $action)
 {
    return ($action -in $script:validRuleActions)
 }

 # GetInvalidRules - returns a string of rule IDs and actions that are NOT
 # supported for this platform or requested an invalid action
 # Format: "d1e49aac-8f56-4280-b9ba-993a6d77406c=1 d4f940ab-401b-4efc-aadc-ad5f3c50688a=2"
 function GetInvalidRules([hashtable] $rulesInfoMap)
 {
    $invalidRules = ""

    foreach($ruleId in $rulesInfoMap.keys)
    {
        if(!(isRuleIdSupported $ruleId) -or !(isActionSupported $rulesInfoMap[$ruleId]))
        {
            $invalidRules += " $ruleId=" + $rulesInfoMap[$ruleId]
        }
    }

    return $invalidRules
 }

 function GetCurrentValueFromReg([string]$settingSuffix, [string]$regKeyPath)
 {
    switch ($settingsNamesMapping.$settingSuffix)
    {
        $settingsNamesMapping.EnableControlledFolderAccess {
            $currentGPRegItem = (GetSingleCurrentValueFromRegistry -registryPath (Join-Path $script:ExploitGuardRegPath $script:regKeySuffixPerSetting.($settingsNamesMapping.$settingSuffix)) -valueName "EnableControlledFolderAccess")
            break
        }
        {($_ -eq $settingsNamesMapping.AttackSurfaceReductionRules) -or ($_ -in $script:savedAsSeparateValuesInRegistry )} {
            #cannot use -AdditionalChildPath because it is not supported on older PS versions
            $GPKeyPath = (GetRegPathToKey ([registryType]::GP) $settingSuffix)
            $currentGPRegItem = (Get-Item $GPKeyPath).Property
            break
        }
        Default {
            $currentGPRegItem = $null
        }
    }

    return $currentGPRegItem
 }

 # ClearCurrentASRContent - clear content of registry (on $regKeyPath) by removing $value received
 function ClearCurrentASRContent([string] $settingSuffix, [string] $regKeyPath, [string] $value)
 {
    LogInfo (-join("ASR::ClearCurrentASRContent for ", $settingSuffix ," in:", $regKeyPath, " with value:",$value))

    $removedValuesArr = $null
    if($value) 
    {
        $removedValuesArr = $value.Split("|")
    }

    if($settingsNamesMapping.$settingSuffix -eq $settingsNamesMapping.AttackSurfaceReductionRules)
    {   
        if($value)
        {
            $removedValuesArr = $script:conflictUtilsModule."GetValuesToRemove"($value)
        }
    }

    if(!$removedValuesArr -or ($removedValuesArr.Count -eq 0)) #value was null
    {
        # if no value given (old senseCM version), clear will be based on content in GP regisry
        $removedValuesArr = GetCurrentValueFromReg $local:settingSuffix $local:regKeyPath
        LogInfo (-join("ASR: remove ", $local:settingSuffix, ", no value, deleting all content from Reg:",$local:regKeyPath))
    }
        
    if($local:removedValuesArr.Count -gt 0)
    {
        LogInfo (-join("ASR: removing [", ($local:removedValuesArr -join('|')), "]"))
        Remove-ItemProperty -Path $regKeyPath -Name $local:removedValuesArr -ErrorAction SilentlyContinue 
    }
 }

 # GetGroupFromRules - function uses requestedValues data struct to 
 # get the common group for all the rules recieved.
 # if no such group is found, $null is returned
 function GetGroupFromRules ([string[]] $rulesArr)
 {
    $rulesInfo = $script:requestedValues[$settingsNamesMapping.AttackSurfaceReductionRules]
    
    $groups = @()
    foreach($ruleId in $rulesArr)
    {
        $singleRuleInfo = $rulesInfo[$ruleId]
        $ruleGroups = $singleRuleInfo[([infoIndex]::groups)]
        if($groups.Count -eq 0)
        {
            #first rule found - take into account all groups
            $groups = $ruleGroups
        } 
        else 
        {
            #leave only the group ids that exists in both
            $groups = ($groups | ?{$ruleGroups -contains $_}) 
        }
    }
    
    # the rules recieved have more than 1 group in common 
    # remove the groups that contain other rules as well
    if($groups.Count -gt 1) 
    {
        LogInfo (-join("ASR GetGroupFromRules common groups: ", ($groups -join "|")))
        foreach($singleRule in $rulesInfo.Keys)
        {
            $singleRuleInfo = $rulesInfo[$singleRule]
            $singleRuleGroups = $singleRuleInfo[([infoIndex]::groups)]
            $commonGroups = ($groups | ?{$singleRuleGroups -contains $_})

            if( ($commonGroups -ne $null) -and !($rulesArr -contains $singleRule))
            {
                $groups = ($groups | ?{!($commonGroups -contains $_)}) 
                if($groups.Count -eq 1) 
                {
                    #if groups is filtered enough (1 group) - stop
                    break
                }
            }
        }
    }
    
    LogInfo (-join("ASR GetGroupFromRules identifying groups: ", ($groups -join "|")))
    
    #at this point we either have only 1 group or several groups with the same set of rules 
    # (could be if order is not the same so PolicyEnforcer does not recognize that they are the same), 
    # and then they must have the same result (conflict/no conflict).
    return $groups
 }

 function GetNameSuffix([string] $fullSettingName)
 {
     return $fullSettingName.Replace($script:settingNamePrefix, "")
 }

#RetrieveTestInfo - returns module info for the TestRunner. Returning here any variable and struct needed for running the tests for this module
function RetrieveTestInfo() {
    return $settingsNamesMapping, (-join($script:RegPrefix,$script:ExploitGuardRegPath)), $script:settingNamePrefix, $script:regKeySuffixPerSetting
}
# SIG # Begin signature block
# MIIntwYJKoZIhvcNAQcCoIInqDCCJ6QCAQExDzANBglghkgBZQMEAgEFADB5Bgor
# BgEEAYI3AgEEoGswaTA0BgorBgEEAYI3AgEeMCYCAwEAAAQQH8w7YFlLCE63JNLG
# KX7zUQIBAAIBAAIBAAIBAAIBADAxMA0GCWCGSAFlAwQCAQUABCCqVlg77ZSi/E95
# BBDkj5OT07Y7kDa/l0PeSve294ULpaCCDZcwggYVMIID/aADAgECAhMzAAADcaWg
# nFyRHuruAAAAAANxMA0GCSqGSIb3DQEBCwUAMH4xCzAJBgNVBAYTAlVTMRMwEQYD
# VQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNpZ25p
# bmcgUENBIDIwMTEwHhcNMjMwNTExMTg0MTEzWhcNMjQwNTA4MTg0MTEzWjCBlDEL
# MAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1v
# bmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjE+MDwGA1UEAxM1TWlj
# cm9zb2Z0IFdpbmRvd3MgRGVmZW5kZXIgQWR2YW5jZWQgVGhyZWF0IFByb3RlY3Rp
# b24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDQl+fOHFGkNfDHrcv3
# g8shIsUp58WmSbtNm4uu7Sw75Wgsgo5p2yYGGqiRWbu0eIKhAJ1bRp9Qa7vAjMEw
# d0wf9wlLjDNAITzDsokwoCl0rOgWxzYuGSusZq4e/liClIpY5fuxrpsEnsssR1o6
# jITkdJQ6ZkWsTFETxLLOEGxe9BkTkXZaBOHlhTt5QIq6UB6j8zNN/Vo4rFmq751X
# lQLA1jOnNi5NOS3/11TW16ZCo9XKG22zdNsJCUckT/QH2eeQNgpuyfIvP06o+RpP
# Uo7P4xdc4c74S3HKfRiNuqbs3+xvb22GtXIz3cKFDhILpNKN/SC8qLjizq0PxwL1
# lr1ZAgMBAAGjggFzMIIBbzAfBgNVHSUEGDAWBggrBgEFBQcDAwYKKwYBBAGCN0wv
# ATAdBgNVHQ4EFgQUcR5wDY+Mh5A/I7scm54A8ODeSVcwRQYDVR0RBD4wPKQ6MDgx
# HjAcBgNVBAsTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEWMBQGA1UEBRMNNDUxODk0
# KzUwMTAwNTAfBgNVHSMEGDAWgBRIbmTlUAXTgqoXNzcitW2oynUClTBUBgNVHR8E
# TTBLMEmgR6BFhkNodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpb3BzL2NybC9N
# aWNDb2RTaWdQQ0EyMDExXzIwMTEtMDctMDguY3JsMGEGCCsGAQUFBwEBBFUwUzBR
# BggrBgEFBQcwAoZFaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraW9wcy9jZXJ0
# cy9NaWNDb2RTaWdQQ0EyMDExXzIwMTEtMDctMDguY3J0MAwGA1UdEwEB/wQCMAAw
# DQYJKoZIhvcNAQELBQADggIBAKYhzPfo4pWE2MYg2W1VnHJ7NZ/nwEpGYEz7Qhqc
# bPj2sJFFhnUzYHOAyXz4vP0+ERhFw9lrqLvoyj8s6xQU11QPQVItopahH65USGWh
# UcyrXFBVSx7MG3Dwi+NT6rxq9MjY3KLB1hZPF7hgtOjWgtjXgeJOsy3IjPiRkDwX
# 5b5Pg6o8Xpyt84mpvAPsDKO7yJGJq8V528deIYh+0DcELvUZTtl4Kz8ElT4BxF6i
# XxXkuqpCWO/rwEa6HrLNV2udND/AImVio09bbZXOWBcb9uTTNezYX5fZvinIX7hl
# UHHIDaVUqqwTFtac8q7BLE02T0P7NAeBtayujIQFzGNuwLN302UmUWAyc4LynqZa
# h2OmZWcMODUCuHIMP3pk54XjWt0YYEWPFAbA7T7j8wjby/vPXdG93LybID/yjGDb
# U1t5xMoEpN5rvSUM1U42iJ1tF268/nE2/8nKqUpdJXH/MaXxYU3FnbshmIe1xfxF
# NjeMYN4jg+5gjz5SBzHAhT1X9QWXNOhAXxib3wcchlsxFYtSNihHLEJ2GuckRuPi
# 47VWi/3hTTnl88S8r3jsbWLatGbQxcbrsF4sglAvFD9SQSXLGUsE5wz/1p/Gb1Fh
# qfNBc6E4AWVj9+A1KezSnMBAw8gJk3fiRkdxfSfpLOkHvgA7qsOmQ7qwjs05kWZf
# 055CMIIHejCCBWKgAwIBAgIKYQ6Q0gAAAAAAAzANBgkqhkiG9w0BAQsFADCBiDEL
# MAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1v
# bmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEyMDAGA1UEAxMpTWlj
# cm9zb2Z0IFJvb3QgQ2VydGlmaWNhdGUgQXV0aG9yaXR5IDIwMTEwHhcNMTEwNzA4
# MjA1OTA5WhcNMjYwNzA4MjEwOTA5WjB+MQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMSgwJgYDVQQDEx9NaWNyb3NvZnQgQ29kZSBTaWduaW5nIFBD
# QSAyMDExMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEAq/D6chAcLq3Y
# bqqCEE00uvK2WCGfQhsqa+laUKq4BjgaBEm6f8MMHt03a8YS2AvwOMKZBrDIOdUB
# FDFC04kNeWSHfpRgJGyvnkmc6Whe0t+bU7IKLMOv2akrrnoJr9eWWcpgGgXpZnbo
# MlImEi/nqwhQz7NEt13YxC4Ddato88tt8zpcoRb0RrrgOGSsbmQ1eKagYw8t00CT
# +OPeBw3VXHmlSSnnDb6gE3e+lD3v++MrWhAfTVYoonpy4BI6t0le2O3tQ5GD2Xuy
# e4Yb2T6xjF3oiU+EGvKhL1nkkDstrjNYxbc+/jLTswM9sbKvkjh+0p2ALPVOVpEh
# NSXDOW5kf1O6nA+tGSOEy/S6A4aN91/w0FK/jJSHvMAhdCVfGCi2zCcoOCWYOUo2
# z3yxkq4cI6epZuxhH2rhKEmdX4jiJV3TIUs+UsS1Vz8kA/DRelsv1SPjcF0PUUZ3
# s/gA4bysAoJf28AVs70b1FVL5zmhD+kjSbwYuER8ReTBw3J64HLnJN+/RpnF78Ic
# V9uDjexNSTCnq47f7Fufr/zdsGbiwZeBe+3W7UvnSSmnEyimp31ngOaKYnhfsi+E
# 11ecXL93KCjx7W3DKI8sj0A3T8HhhUSJxAlMxdSlQy90lfdu+HggWCwTXWCVmj5P
# M4TasIgX3p5O9JawvEagbJjS4NaIjAsCAwEAAaOCAe0wggHpMBAGCSsGAQQBgjcV
# AQQDAgEAMB0GA1UdDgQWBBRIbmTlUAXTgqoXNzcitW2oynUClTAZBgkrBgEEAYI3
# FAIEDB4KAFMAdQBiAEMAQTALBgNVHQ8EBAMCAYYwDwYDVR0TAQH/BAUwAwEB/zAf
# BgNVHSMEGDAWgBRyLToCMZBDuRQFTuHqp8cx0SOJNDBaBgNVHR8EUzBRME+gTaBL
# hklodHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNS
# b29DZXJBdXQyMDExXzIwMTFfMDNfMjIuY3JsMF4GCCsGAQUFBwEBBFIwUDBOBggr
# BgEFBQcwAoZCaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNS
# b29DZXJBdXQyMDExXzIwMTFfMDNfMjIuY3J0MIGfBgNVHSAEgZcwgZQwgZEGCSsG
# AQQBgjcuAzCBgzA/BggrBgEFBQcCARYzaHR0cDovL3d3dy5taWNyb3NvZnQuY29t
# L3BraW9wcy9kb2NzL3ByaW1hcnljcHMuaHRtMEAGCCsGAQUFBwICMDQeMiAdAEwA
# ZQBnAGEAbABfAHAAbwBsAGkAYwB5AF8AcwB0AGEAdABlAG0AZQBuAHQALiAdMA0G
# CSqGSIb3DQEBCwUAA4ICAQBn8oalmOBUeRou09h0ZyKbC5YR4WOSmUKWfdJ5DJDB
# ZV8uLD74w3LRbYP+vj/oCso7v0epo/Np22O/IjWll11lhJB9i0ZQVdgMknzSGksc
# 8zxCi1LQsP1r4z4HLimb5j0bpdS1HXeUOeLpZMlEPXh6I/MTfaaQdION9MsmAkYq
# wooQu6SpBQyb7Wj6aC6VoCo/KmtYSWMfCWluWpiW5IP0wI/zRive/DvQvTXvbiWu
# 5a8n7dDd8w6vmSiXmE0OPQvyCInWH8MyGOLwxS3OW560STkKxgrCxq2u5bLZ2xWI
# UUVYODJxJxp/sfQn+N4sOiBpmLJZiWhub6e3dMNABQamASooPoI/E01mC8CzTfXh
# j38cbxV9Rad25UAqZaPDXVJihsMdYzaXht/a8/jyFqGaJ+HNpZfQ7l1jQeNbB5yH
# PgZ3BtEGsXUfFL5hYbXw3MYbBL7fQccOKO7eZS/sl/ahXJbYANahRr1Z85elCUtI
# EJmAH9AAKcWxm6U/RXceNcbSoqKfenoi+kiVH6v7RyOA9Z74v2u3S5fi63V4Guzq
# N5l5GEv/1rMjaHXmr/r8i+sLgOppO6/8MO0ETI7f33VtY5E90Z1WTk+/gFcioXgR
# MiF670EKsT/7qMykXcGhiJtXcVZOSEXAQsmbdlsKgEhr/Xmfwb1tbWrJUnMTDXpQ
# zTGCGXYwghlyAgEBMIGVMH4xCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5n
# dG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9y
# YXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNpZ25pbmcgUENBIDIwMTEC
# EzMAAANxpaCcXJEe6u4AAAAAA3EwDQYJYIZIAWUDBAIBBQCgga4wGQYJKoZIhvcN
# AQkDMQwGCisGAQQBgjcCAQQwHAYKKwYBBAGCNwIBCzEOMAwGCisGAQQBgjcCARUw
# LwYJKoZIhvcNAQkEMSIEIEZ/zNBn0It102NLDEu5woLKRFIdQCmOSvjqrRij88Pj
# MEIGCisGAQQBgjcCAQwxNDAyoBSAEgBNAGkAYwByAG8AcwBvAGYAdKEagBhodHRw
# Oi8vd3d3Lm1pY3Jvc29mdC5jb20wDQYJKoZIhvcNAQEBBQAEggEAx4F+7JrZQ9kT
# b5wtZmlWG2KYyfzM1geg7zCwhllhHNPCOzyr1qPHdZEdO/3TKb47UejvTd1phTcB
# mMukgg4D62Hf3IV8iYzChtObhJvY9dkXrWNDJ7VfbZLoP10UfZPBerERzVUOwWbg
# FEn1z6Vt1V+eA9I5Ej8R566fssywMhY/ibW/YLfHyC5Je8phpebXeUB8Vmtui/7E
# JwTPHKFv7JqnuGaGX4rtHYNhkm4/tZeUidWbWNG9b6hHjFj3oXgACt6kU8iPSdS5
# QpabQCkkfo322wdtrV85+gs9K4zi+ntv99JsNo1NuNoSwHl8wx3n6L3KLyyUst3W
# L2M6GY0PEaGCFwAwghb8BgorBgEEAYI3AwMBMYIW7DCCFugGCSqGSIb3DQEHAqCC
# FtkwghbVAgEDMQ8wDQYJYIZIAWUDBAIBBQAwggFRBgsqhkiG9w0BCRABBKCCAUAE
# ggE8MIIBOAIBAQYKKwYBBAGEWQoDATAxMA0GCWCGSAFlAwQCAQUABCDKK7V1rR0Z
# NiGcSlIg9nqbzjdkrqw/WNTXYtE+puQ23AIGZItYsGUuGBMyMDIzMDcxMTE3NDcw
# MC43NTlaMASAAgH0oIHQpIHNMIHKMQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2Fz
# aGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENv
# cnBvcmF0aW9uMSUwIwYDVQQLExxNaWNyb3NvZnQgQW1lcmljYSBPcGVyYXRpb25z
# MSYwJAYDVQQLEx1UaGFsZXMgVFNTIEVTTjpBRTJDLUUzMkItMUFGQzElMCMGA1UE
# AxMcTWljcm9zb2Z0IFRpbWUtU3RhbXAgU2VydmljZaCCEVcwggcMMIIE9KADAgEC
# AhMzAAABv99uuQQVUihYAAEAAAG/MA0GCSqGSIb3DQEBCwUAMHwxCzAJBgNVBAYT
# AlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYD
# VQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xJjAkBgNVBAMTHU1pY3Jvc29mdCBU
# aW1lLVN0YW1wIFBDQSAyMDEwMB4XDTIyMTEwNDE5MDEyNFoXDTI0MDIwMjE5MDEy
# NFowgcoxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQH
# EwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xJTAjBgNV
# BAsTHE1pY3Jvc29mdCBBbWVyaWNhIE9wZXJhdGlvbnMxJjAkBgNVBAsTHVRoYWxl
# cyBUU1MgRVNOOkFFMkMtRTMyQi0xQUZDMSUwIwYDVQQDExxNaWNyb3NvZnQgVGlt
# ZS1TdGFtcCBTZXJ2aWNlMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEA
# uExh0n1UxKMzBvkPHer47nryD4UK2GVy1X6bOVC+hLVhDlsIWQ1uX/9a8IRI3zXo
# /y1oTDuj+rJHyX4OZQn42E0iu7x6swPvM34zIOSPn8lgnWzGEAsRtz9zBrLW9+4w
# /YhWlXI8hvc7ovqupuL3TXte8BbmNOUDSL+Ou2bBfObGzsH3yY/BELvqwO13KZ9Z
# 1OxKacnqq1u9E9Rhai90STog22lR2MVRSx55FHi/emnZA/IKvsAtEH2K6JmgOyQ7
# /mDQrWNEA5roUjhQqLQw1/3wz/CIvc9+FPxX2dxR0nvvYe5VLqv8Q99cOkO6z6V4
# stGDyFDuO8CwtiSvCC3QrOOugAl33aPD9YZswywWRk+YGyLI+Fw+kCCUY6h1qOjT
# j5glz0esmds3ue45WaI2hI9usForM8gy//5tDZXj0KKU1BxA04xpfEy91RZUbc6p
# dAvEkpYrN2jlpXhMvTD7pgdYyxkVSaWZv7kWp5y9NjWP/CTDGXTC6DWiGcXwPQO6
# 6QdVNWxuiGdpfPaEUnWXcKnDVua1khBAxO4m9wg/1qM6f7HwXf/pHifMej+qB7SU
# ZOiJScX+1HmffmZRAFiJXS0qUDk0ZAZW3oX2xLyl0044eHI7Y95GPaw8OlSTeNiN
# AKl+MyH5OaifsUuyVHOf4rsrE+ZyAuS9e9ERqu5H/10CAwEAAaOCATYwggEyMB0G
# A1UdDgQWBBRVAolUT3eV3wK/+Luf/wawCPMYpzAfBgNVHSMEGDAWgBSfpxVdAF5i
# XYP05dJlpxtTNRnpcjBfBgNVHR8EWDBWMFSgUqBQhk5odHRwOi8vd3d3Lm1pY3Jv
# c29mdC5jb20vcGtpb3BzL2NybC9NaWNyb3NvZnQlMjBUaW1lLVN0YW1wJTIwUENB
# JTIwMjAxMCgxKS5jcmwwbAYIKwYBBQUHAQEEYDBeMFwGCCsGAQUFBzAChlBodHRw
# Oi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpb3BzL2NlcnRzL01pY3Jvc29mdCUyMFRp
# bWUtU3RhbXAlMjBQQ0ElMjAyMDEwKDEpLmNydDAMBgNVHRMBAf8EAjAAMBMGA1Ud
# JQQMMAoGCCsGAQUFBwMIMA0GCSqGSIb3DQEBCwUAA4ICAQAjCREvjT6yXwJYdvkF
# UqTGGh6RizAY+ciuB6UOBUm0yqq5QC+5pCEa9WSMvbUGzxDCEFBgD93gWGnkiyYc
# HCazlgZK+E7WxtI3bP++Fb4RJZiWLo/IC9hX12hCZZwYXIGVzC9BVAcNx/zsFqI/
# 9u8u/bhGjDHPad47C4OQNCHrkNqzGYxb4GQq6Psw6o7cEty3MU3Jd4uzBazaFhPR
# vmBfSn+Ufd6pTNZLgIX9BjrLmZblc/d2LIAurEr5W29WfW5RMRIEZzO9TaMr/zzd
# mW/cV6VdaDTygy5g4O3UXadt1DraUpn5jcD10TVWNnyz/paeleHojrGCCksqexpe
# lMkUsiYP0HX9pFUgNglWU10r1wEzFwZM9aX2Rqq3fFRrN3gu8tCX+H1nKK2AobW1
# vmsKLTH6PyX1LkyvRwTj45a1paeHIR8TGzm3+iY7wpC1MHuzqAqAdDeaIVdVlch8
# 07VJJ4hDive6AiOQCV9MwiUyhf5v4P8jTGof8CqjDb3PnLlNSnFm2BFhMZ35oNTE
# osc37GZHScM83hTN1E481sLYJrrhhcdtcyNB60juMjqGUD6uQ/7DbMvtv93tFj5W
# jxVhMCkkY66EEYgpfFLOCb2ngJJWFuJCIGsCiDfDxGwE4RVYAnoFzoa2OfSqijYg
# 2drdZfpptRRvKxMsAzu3oxkS/TCCB3EwggVZoAMCAQICEzMAAAAVxedrngKbSZkA
# AAAAABUwDQYJKoZIhvcNAQELBQAwgYgxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpX
# YXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQg
# Q29ycG9yYXRpb24xMjAwBgNVBAMTKU1pY3Jvc29mdCBSb290IENlcnRpZmljYXRl
# IEF1dGhvcml0eSAyMDEwMB4XDTIxMDkzMDE4MjIyNVoXDTMwMDkzMDE4MzIyNVow
# fDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1Jl
# ZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEmMCQGA1UEAxMd
# TWljcm9zb2Z0IFRpbWUtU3RhbXAgUENBIDIwMTAwggIiMA0GCSqGSIb3DQEBAQUA
# A4ICDwAwggIKAoICAQDk4aZM57RyIQt5osvXJHm9DtWC0/3unAcH0qlsTnXIyjVX
# 9gF/bErg4r25PhdgM/9cT8dm95VTcVrifkpa/rg2Z4VGIwy1jRPPdzLAEBjoYH1q
# UoNEt6aORmsHFPPFdvWGUNzBRMhxXFExN6AKOG6N7dcP2CZTfDlhAnrEqv1yaa8d
# q6z2Nr41JmTamDu6GnszrYBbfowQHJ1S/rboYiXcag/PXfT+jlPP1uyFVk3v3byN
# pOORj7I5LFGc6XBpDco2LXCOMcg1KL3jtIckw+DJj361VI/c+gVVmG1oO5pGve2k
# rnopN6zL64NF50ZuyjLVwIYwXE8s4mKyzbnijYjklqwBSru+cakXW2dg3viSkR4d
# Pf0gz3N9QZpGdc3EXzTdEonW/aUgfX782Z5F37ZyL9t9X4C626p+Nuw2TPYrbqgS
# Uei/BQOj0XOmTTd0lBw0gg/wEPK3Rxjtp+iZfD9M269ewvPV2HM9Q07BMzlMjgK8
# QmguEOqEUUbi0b1qGFphAXPKZ6Je1yh2AuIzGHLXpyDwwvoSCtdjbwzJNmSLW6Cm
# gyFdXzB0kZSU2LlQ+QuJYfM2BjUYhEfb3BvR/bLUHMVr9lxSUV0S2yW6r1AFemzF
# ER1y7435UsSFF5PAPBXbGjfHCBUYP3irRbb1Hode2o+eFnJpxq57t7c+auIurQID
# AQABo4IB3TCCAdkwEgYJKwYBBAGCNxUBBAUCAwEAATAjBgkrBgEEAYI3FQIEFgQU
# KqdS/mTEmr6CkTxGNSnPEP8vBO4wHQYDVR0OBBYEFJ+nFV0AXmJdg/Tl0mWnG1M1
# GelyMFwGA1UdIARVMFMwUQYMKwYBBAGCN0yDfQEBMEEwPwYIKwYBBQUHAgEWM2h0
# dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2lvcHMvRG9jcy9SZXBvc2l0b3J5Lmh0
# bTATBgNVHSUEDDAKBggrBgEFBQcDCDAZBgkrBgEEAYI3FAIEDB4KAFMAdQBiAEMA
# QTALBgNVHQ8EBAMCAYYwDwYDVR0TAQH/BAUwAwEB/zAfBgNVHSMEGDAWgBTV9lbL
# j+iiXGJo0T2UkFvXzpoYxDBWBgNVHR8ETzBNMEugSaBHhkVodHRwOi8vY3JsLm1p
# Y3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNSb29DZXJBdXRfMjAxMC0w
# Ni0yMy5jcmwwWgYIKwYBBQUHAQEETjBMMEoGCCsGAQUFBzAChj5odHRwOi8vd3d3
# Lm1pY3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY1Jvb0NlckF1dF8yMDEwLTA2LTIz
# LmNydDANBgkqhkiG9w0BAQsFAAOCAgEAnVV9/Cqt4SwfZwExJFvhnnJL/Klv6lwU
# tj5OR2R4sQaTlz0xM7U518JxNj/aZGx80HU5bbsPMeTCj/ts0aGUGCLu6WZnOlNN
# 3Zi6th542DYunKmCVgADsAW+iehp4LoJ7nvfam++Kctu2D9IdQHZGN5tggz1bSNU
# 5HhTdSRXud2f8449xvNo32X2pFaq95W2KFUn0CS9QKC/GbYSEhFdPSfgQJY4rPf5
# KYnDvBewVIVCs/wMnosZiefwC2qBwoEZQhlSdYo2wh3DYXMuLGt7bj8sCXgU6ZGy
# qVvfSaN0DLzskYDSPeZKPmY7T7uG+jIa2Zb0j/aRAfbOxnT99kxybxCrdTDFNLB6
# 2FD+CljdQDzHVG2dY3RILLFORy3BFARxv2T5JL5zbcqOCb2zAVdJVGTZc9d/HltE
# AY5aGZFrDZ+kKNxnGSgkujhLmm77IVRrakURR6nxt67I6IleT53S0Ex2tVdUCbFp
# AUR+fKFhbHP+CrvsQWY9af3LwUFJfn6Tvsv4O+S3Fb+0zj6lMVGEvL8CwYKiexcd
# FYmNcP7ntdAoGokLjzbaukz5m/8K6TT4JDVnK+ANuOaMmdbhIurwJ0I9JZTmdHRb
# atGePu1+oDEzfbzL6Xu/OHBE0ZDxyKs6ijoIYn/ZcGNTTY3ugm2lBRDBcQZqELQd
# VTNYs6FwZvKhggLOMIICNwIBATCB+KGB0KSBzTCByjELMAkGA1UEBhMCVVMxEzAR
# BgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1p
# Y3Jvc29mdCBDb3Jwb3JhdGlvbjElMCMGA1UECxMcTWljcm9zb2Z0IEFtZXJpY2Eg
# T3BlcmF0aW9uczEmMCQGA1UECxMdVGhhbGVzIFRTUyBFU046QUUyQy1FMzJCLTFB
# RkMxJTAjBgNVBAMTHE1pY3Jvc29mdCBUaW1lLVN0YW1wIFNlcnZpY2WiIwoBATAH
# BgUrDgMCGgMVADgEd+JNrp4dpvFKMZi91txbfic3oIGDMIGApH4wfDELMAkGA1UE
# BhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAc
# BgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEmMCQGA1UEAxMdTWljcm9zb2Z0
# IFRpbWUtU3RhbXAgUENBIDIwMTAwDQYJKoZIhvcNAQEFBQACBQDoV3TMMCIYDzIw
# MjMwNzExMTQyNzI0WhgPMjAyMzA3MTIxNDI3MjRaMHcwPQYKKwYBBAGEWQoEATEv
# MC0wCgIFAOhXdMwCAQAwCgIBAAICCrcCAf8wBwIBAAICET8wCgIFAOhYxkwCAQAw
# NgYKKwYBBAGEWQoEAjEoMCYwDAYKKwYBBAGEWQoDAqAKMAgCAQACAwehIKEKMAgC
# AQACAwGGoDANBgkqhkiG9w0BAQUFAAOBgQBGqBtOF/IvW3fOBHvqlAvYkIc5B/mR
# V2NdLmDokXvlmhv6nrgjfGNueSjiM1fn3TIq8p9LYAwvgZDcfVdK8tyTE5ara4KT
# iu44K7RhD7aK8iRiC4HwvYkRyDDVpPZgb7gB5KIhgyxKLzzrrqKDo2MMxHRKivMC
# Gs0X+3NLj4eBBTGCBA0wggQJAgEBMIGTMHwxCzAJBgNVBAYTAlVTMRMwEQYDVQQI
# EwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3Nv
# ZnQgQ29ycG9yYXRpb24xJjAkBgNVBAMTHU1pY3Jvc29mdCBUaW1lLVN0YW1wIFBD
# QSAyMDEwAhMzAAABv99uuQQVUihYAAEAAAG/MA0GCWCGSAFlAwQCAQUAoIIBSjAa
# BgkqhkiG9w0BCQMxDQYLKoZIhvcNAQkQAQQwLwYJKoZIhvcNAQkEMSIEIGxWUhBD
# grLiCAjoCxSkDV31CYFqs8V94ln+dpLJPOj4MIH6BgsqhkiG9w0BCRACLzGB6jCB
# 5zCB5DCBvQQg/Q4tRz63EiRj4K+19yNUwogBIOsp44CIuBfnZHCvBa4wgZgwgYCk
# fjB8MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMH
# UmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSYwJAYDVQQD
# Ex1NaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EgMjAxMAITMwAAAb/fbrkEFVIoWAAB
# AAABvzAiBCB37GVV1xAgkvmPsThf8SWHT1eqjb5agEvSFESjyfuMBjANBgkqhkiG
# 9w0BAQsFAASCAgBrOt/agY+jWyVgVVNyWiAPpu5FM1KurPz8dZsrD4Sng3ACVPlH
# I3RSFE283t7WpYgH2hQ/HPvISHGHdkOkGViFcuLNgZZxOC9ila2YONq6dg9j/FVh
# 86d1mbMwuu6pEb02KWDUXyXuHiXdSnZ2n5y/LgVivs7JU6ZysixjcyqSee97NXfP
# WESbGBwseP6aoOroYVzwZlbc4kWf7pAV1wyedoSUlvjAgSB0zJRQP+wbi868JCxx
# w//najVNNBGBj9uHFzF8jR1IiHR7oTCtILTmEeHZQaNoSDPeo8XZqA3/EPDZbZay
# 19qoRn6H1KivqYRSfn8O/hXiSTxa/JdQh1Gj/Gu/khdTeWl9cZjqdmYd5Z1d2DtU
# SHaFHH5Iq1cqHLzpx2Hc8jkCdhTCozGaNTIXxSXIKGR3iMsxglTzmAQaTd1WDote
# xTN0L11X7AnLYrcYC+jgQh2WjdRyaDhfgi9I8iN/ynJvTVsU//weyp44nDVhUvnj
# IM+xbE13YhIBGSDz+rEP9AQbhhTeDU4HiHHVoRKip3lgLZvy8sMm8fDjWXyKtk69
# eDsi+I6RS1Zm5ypbh9MnIK+WVoxI4xhHaXig/43znd55HD1t39mZKb3xyLLuVj8X
# WxqZNSfsYeyFtmPaO4WPpo4PcFUmX6gE3Y367b+ulDHz5eyWw9X95vV8cw==
# SIG # End signature block
